# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple, Optional, List, Dict, Any, Callable, Union
import torch
import torch.distributed.pipelining
from torch.distributed.pipelining import (
    PipelineStage as OriginalPipelineStage,
    Schedule1F1B as OriginalSchedule1F1B,
    ScheduleGPipe as OriginalScheduleGPipe,
)
from torch.distributed.pipelining.stage import (
    _RecvInfo,
    _RootArgPlaceholder,
    _make_tensor_from_meta,
)

from cosmos_rl.utils.logging import logger


# Common functions for dynamic shape support
def clear_stage(self):
    self._stage_initialized = False
    self._stage.inputs_meta = None
    self._stage._outputs_meta = None


def infer_seq_dim(self, position_ids_shape):
    # Find the dimension index where the current position_ids shape differs from the initial shape
    diff = [
        i
        for i, (a, b) in enumerate(
            zip(position_ids_shape, self.init_position_ids_shape)
        )
        if a != b
    ]
    if len(diff) > 1 and not self.clear_stage_enabled:
        logger.warning(
            "Only one dimension can be different between the current and the initial position ids shape.\
                         Set clear_stage_enabled to True."
        )
        self.clear_stage_enabled = True
        return

    if len(diff) == 1:
        position_ids_seq_dim = diff[0]
        self.position_ids_seq_dim = position_ids_seq_dim
        prev_seq_len = self.init_position_ids_shape[position_ids_seq_dim]
        input_prev_seq_len = prev_seq_len // self.seq_len_multiple
        output_prev_seq_len = (
            input_prev_seq_len if not self._stage.is_last else prev_seq_len
        )

        logger.info(f"Inferred position ids seq dim: {self.position_ids_seq_dim}")
        for input_meta in self._stage.inputs_meta:
            input_meta_shape = input_meta.shape
            logger.info(f"Input meta shape: {input_meta_shape} {input_prev_seq_len=}")
            for i, dim in enumerate(input_meta_shape):
                if dim == input_prev_seq_len:
                    self.input_seq_dim = i
                    logger.info(f"Inferred input meta seq dim: {self.input_seq_dim}")
                    break
            if self.input_seq_dim is None:
                logger.warning(
                    f"Seq dim is not inferred for input meta {input_meta}, set clear_stage_enabled to True."
                )
                self.clear_stage_enabled = True
                break
        for output_meta in self._stage._outputs_meta:
            output_meta_shape = output_meta.shape
            logger.info(
                f"Output meta shape: {output_meta_shape} {output_prev_seq_len=}"
            )
            for i, dim in enumerate(output_meta_shape):
                if dim == output_prev_seq_len:
                    self.output_seq_dim = i
                    logger.info(f"Inferred output meta seq dim: {self.output_seq_dim}")
                    break
            if self.output_seq_dim is None:
                logger.warning(
                    f"Seq dim is not inferred for output meta {output_meta}, set clear_stage_enabled to True."
                )
                self.clear_stage_enabled = True


def update_stage(self, position_ids_shape):
    # TODO: support multiple inputs/outputs meta update
    inputs_meta_len = len(self._stage.inputs_meta)
    outputs_meta_len = len(self._stage._outputs_meta)
    if (inputs_meta_len > 1 or outputs_meta_len > 1) and not self.clear_stage_enabled:
        logger.warning(f"Default pipeline parallelism dynamic shape mode currently only supports at most 1 input meta (got {inputs_meta_len}) \
                         or 1 output meta (got {outputs_meta_len}). Set clear_stage_enabled to True.")
        self.clear_stage_enabled = True

    if (
        self.input_seq_dim is None
        and self.output_seq_dim is None
        and not self.clear_stage_enabled
    ):
        self._infer_seq_dim(position_ids_shape)

    # In case the position_ids seq dim is not inferred, just skip the update
    # e.g. both prev batch and current batch have same seq len
    if self.position_ids_seq_dim is None:
        return

    # clear the stage if clear_stage_enabled is True
    if self.clear_stage_enabled:
        self._clear_stage()
        return

    new_seq_len = position_ids_shape[self.position_ids_seq_dim]
    input_new_seq_len = new_seq_len // self.seq_len_multiple
    output_new_seq_len = input_new_seq_len if not self._stage.is_last else new_seq_len
    logger.debug(
        f"Updating stage with {new_seq_len=} {input_new_seq_len=} {output_new_seq_len=}"
    )
    if inputs_meta_len:
        new_inputs_meta = tuple()
        for input_meta in self._stage.inputs_meta:
            input_meta_shape = list(input_meta.shape)
            if input_meta_shape[self.input_seq_dim] != input_new_seq_len:
                input_meta_shape[self.input_seq_dim] = input_new_seq_len
                new_input_meta = torch.empty(
                    input_meta_shape, device=input_meta.device, dtype=input_meta.dtype
                )
            else:
                new_input_meta = input_meta
            new_inputs_meta += (new_input_meta,)
        self._stage.inputs_meta = new_inputs_meta
        logger.debug(f"Updated inputs meta: {self._stage.inputs_meta}")

    if outputs_meta_len:
        new_outputs_meta = tuple()
        for output_meta in self._stage._outputs_meta:
            output_meta_shape = list(output_meta.shape)
            if output_meta_shape[self.output_seq_dim] != output_new_seq_len:
                output_meta_shape[self.output_seq_dim] = output_new_seq_len
                new_output_meta = torch.empty(
                    output_meta_shape,
                    device=output_meta.device,
                    dtype=output_meta.dtype,
                )
            else:
                new_output_meta = output_meta
            new_outputs_meta += (new_output_meta,)
        self._stage._outputs_meta = new_outputs_meta
        logger.debug(f"Updated outputs meta: {self._stage._outputs_meta}")

    self._stage._reconstruct_forward_infra(self._n_microbatches)
    if self._has_backward:
        self._stage._prepare_backward_infra(self._n_microbatches)


def step_func(self, *args, target=None, losses: Optional[List] = None, **kwargs):
    """
    Run one iteration of the pipeline schedule with *whole-batch* input.
    Will chunk the input into microbatches automatically, and go through the
    microbatches according to the schedule implementation.

    args: positional arguments to the model (as in non-pipeline case).
    kwargs: keyword arguments to the model (as in non-pipeline case).
    target: target for the loss function.
    losses: a list to store the losses for each microbatch.
    """

    # Clean per iteration
    self._stage.clear_runtime_states()

    # Split inputs into microbatches
    args_split, kwargs_split = self._split_inputs(args, kwargs)

    # Split target into microbatches
    if target is not None:
        targets_split = list(torch.tensor_split(target, self._n_microbatches))
    else:
        targets_split = None

    if self.pp_dynamic_shape_enabled is None:
        self.pp_dynamic_shape_enabled = kwargs.get("pp_dynamic_shape_enabled", False)

    position_ids = kwargs_split[0].get("position_ids", None)
    if position_ids is None:
        logger.warning("Position ids is None, set clear_stage_enabled to True")
        self.clear_stage_enabled = True
    else:
        if self.pp_dynamic_shape_enabled and self.init_position_ids_shape is None:
            self.init_position_ids_shape = position_ids.shape
        seq_len_multiple = kwargs.get("seq_len_multiple", None)
        if self.pp_dynamic_shape_enabled and self.seq_len_multiple is None:
            assert (
                seq_len_multiple is not None
            ), "Seq len multiple is required for dynamic shape mode"
            self.seq_len_multiple = seq_len_multiple
    # Update stage if it is already initialized by updating input/output metadata
    # to match the new input shapes and types
    if self.pp_dynamic_shape_enabled and self._stage_initialized:
        self._update_stage(position_ids.shape)

    # Run microbatches
    self._step_microbatches(args_split, kwargs_split, targets_split, losses)

    # Return merged results per original format
    if self._stage.is_last:
        return self._merge_outputs(self._stage.output_chunks)
    else:
        return None


class ScheduleGPipe(OriginalScheduleGPipe):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Used for input/output metadata update in dynamic shape mode
        self.input_seq_dim = None
        self.output_seq_dim = None
        self.position_ids_seq_dim = None
        self.init_position_ids_shape = None
        self.seq_len_multiple = None
        self.pp_dynamic_shape_enabled = None
        self.first_call = True
        # If true, clear the stage each step
        self.clear_stage_enabled = False

    # Reset the stage initialized flag for dynamic shape support
    def _clear_stage(self):
        clear_stage(self)

    def _infer_seq_dim(self, position_ids_shape):
        infer_seq_dim(self, position_ids_shape)

    def _update_stage(self, position_ids_shape):
        update_stage(self, position_ids_shape)

    def step(self, *args, target=None, losses: Optional[List] = None, **kwargs):
        # training schedule and validation schedule share the same stage, so we need to clear
        # the stage before the first step
        if self.first_call and (
            len(self._stage.inputs_meta) or len(self._stage._outputs_meta)
        ):
            self._clear_stage()
            self.first_call = False

        return step_func(self, *args, target=target, losses=losses, **kwargs)


class Schedule1F1B(OriginalSchedule1F1B):
    def __init__(self, *args, **kwargs):
        if "scale_grads" not in kwargs and torch.__version__ >= "2.7.0":
            # Loss scale is enabled by default in torch 2.7.0
            # We scale the grads manually in this codebase
            kwargs["scale_grads"] = False
        super().__init__(*args, **kwargs)
        # Used for input/output metadata update in dynamic shape mode
        self.input_seq_dim = None
        self.output_seq_dim = None
        self.position_ids_seq_dim = None
        self.init_position_ids_shape = None
        self.seq_len_multiple = None
        self.pp_dynamic_shape_enabled = None
        # If true, clear the stage each step
        self.clear_stage_enabled = False

    # Reset the stage initialized flag for dynamic shape support
    def _clear_stage(self):
        clear_stage(self)

    def _infer_seq_dim(self, position_ids_shape):
        infer_seq_dim(self, position_ids_shape)

    def _update_stage(self, position_ids_shape):
        update_stage(self, position_ids_shape)

    def step(self, *args, target=None, losses: Optional[List] = None, **kwargs):
        return step_func(self, *args, target=target, losses=losses, **kwargs)


class PipelineStage(OriginalPipelineStage):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if torch.__version__ >= "2.6.0" and torch.__version__ < "2.7.0":
            self.patched = True
        else:
            self.patched = False

    def backward_one_chunk(self, *args, **kwargs):
        if self.patched:
            return self.cosmos_backward_one_chunk(*args, **kwargs)
        else:
            return super().backward_one_chunk(*args, **kwargs)

    def _reconstruct_forward_infra(
        self,
        num_microbatches: int,
    ) -> Tuple[Any, ...]:
        for chunk_id in range(num_microbatches):
            if not self.is_first:
                # We assume that we always receive from stage - 1
                recv_infos = tuple(
                    [
                        _RecvInfo(
                            f"recv_for_{self.stage_index}_from_{self.stage_index - 1}",
                            self.stage_index - 1,
                            _make_tensor_from_meta(inp, self.device),
                        )
                        for inp in self.inputs_meta
                    ]
                )
                # In case there is backward pass, set requires_grad for receive buffers
                if self.has_backward:
                    for r in recv_infos:
                        r.buffer.requires_grad_(True)

                self.args_recv_info[chunk_id] = recv_infos
            else:
                self.args_recv_info[chunk_id] = tuple(
                    [_RootArgPlaceholder(i) for i in self.inputs_meta]
                )

        # Send info during forward for each activation
        # only need the rank that is being sent to
        self.act_send_info: Dict[int, List] = {}

        for idx in range(len(self.get_outputs_meta())):
            # We assume we always send to stage + 1
            if not self.is_last:
                self.act_send_info[idx] = [self.stage_index + 1]
            else:
                self.act_send_info[idx] = []

    def cosmos_backward_one_chunk(
        self,
        bwd_chunk_id: int,
        loss=None,
        full_backward: bool = True,
        last_backward=False,
    ):
        """
        Perform backward pass on the module.
        This should only be called once per microbatch.

        If full_backward is True (the default), the full backward pass including weight and input gradients will be run,
        and it is an error to call `backward_weight_one_chunk` for this bwd_chunk_id.

        If full_backward is False, it is optional that `dw_runner` was provided to the PipelineStage at __init__ time,
        and a subsequent call to `backward_weight_one_chunk` is required to invoke dw_runner and complete the backward.

        last_backward is controlled by the schedule and signals synchronization of gradients across DP groups
        after the last backward.
        """
        self._check_chunk_id(bwd_chunk_id)

        (
            stage_output,
            input_values,
        ) = self.fwd_cache.pop(bwd_chunk_id)

        # Compute backward
        if self.is_last:
            # Last stage computes gradients from loss and has no gradients from
            # next stage
            bwd_kwargs = {
                "stage_output": loss,
                "output_grads": None,
                "input_values": input_values,
            }
        else:
            # Otherwise, receive gradients from next stage
            grads_output = self._retrieve_recv_grads(bwd_chunk_id)
            # If an input to the pipeline requires gradient,
            # `torch.autograd.backward` will accumulate the gradient into the
            # `.grad` field of such input
            bwd_kwargs = {
                "stage_output": stage_output,
                "output_grads": grads_output,
                "input_values": input_values,
            }

        grads_input: Tuple[Optional[torch.Tensor], ...] = ()

        # Custom backward function
        if self.dw_builder:
            # TODO: We may want to change our semantics so we are allowed to ignore
            # the 'dw_builder' and call full_backward directly when it is a full_backward op.
            grads_input, _ = self.backward_maybe_with_nosync(
                "full", bwd_kwargs, last_backward=last_backward
            )
            if full_backward:
                self.dw_builder()()
            else:
                self.dw_runner[bwd_chunk_id] = self.dw_builder()
        else:
            if full_backward:
                grads_input, _ = self.backward_maybe_with_nosync(
                    "full", bwd_kwargs, last_backward=last_backward
                )
            else:
                param_groups: List[Dict[str, Any]] | None = None
                # Skip the backward for the first stage since we will perform the weight update with
                # autograd.backward in backward_weight_one_chunk
                if not self.is_first:
                    if isinstance(bwd_kwargs["stage_output"], torch.Tensor):
                        bwd_kwargs["stage_output"] = (bwd_kwargs["stage_output"],)

                    # perform the partial backwards for the inputs with a custom backward function
                    # when the "stage_ouput" is a loss, then it is a tensor, otherwise it is a tuple of tensors
                    grads_input, param_groups = self.backward_maybe_with_nosync(
                        "input", bwd_kwargs, last_backward=last_backward
                    )

                # TODO: we dont need to save this, add to dw_runner?
                self.backward_state[bwd_chunk_id] = (
                    bwd_kwargs["input_values"],
                    param_groups,
                    bwd_kwargs["stage_output"],
                    bwd_kwargs["output_grads"],
                )
                # Save a placeholder for the dw_runner
                self.dw_runner[bwd_chunk_id] = lambda: None

        self.bwd_cache[bwd_chunk_id] = grads_input

        if self.is_last and not self.is_first:
            # Autograd dependencies:
            #    rest_of_autograd_graph -> stage_output -> loss
            # stage_output is no longer used in the last stage for backward and only needed
            # to return to the user in merge_output_chunks, therefore
            # this should be detached to release autograd graph context and free memory earlier
            for t in stage_output:
                if not t._is_view():
                    t.detach_()


def patch_fsdp_foreach_reduce():
    from torch.distributed.fsdp._fully_shard._fsdp_param import FSDPParam
    from torch.distributed.device_mesh import _get_device_handle
    from torch.distributed.fsdp._fully_shard._fsdp_collectives import (
        foreach_reduce_scatter_copy_in,
        _div_if_needed,
    )
    import torch.distributed as dist
    from torch.distributed.distributed_c10d import ReduceOp
    from torch.distributed.tensor import DTensor

    def _get_gradient_divide_factors(
        reduce_scatter_group: dist.ProcessGroup,
        all_reduce_group: Optional[dist.ProcessGroup],
        reduce_dtype: torch.dtype,
        device_type: str = "",
    ) -> Union[tuple[None, None], tuple[float, float]]:
        # For fp32/bf16, we do not need to worry about overflow/underflow, so we
        # use NCCL's built-in division to avoid separate div kernels
        if reduce_dtype in (torch.float32, torch.bfloat16) and device_type != "mtia":
            return None, None
        data_parallel_size = reduce_scatter_group.size()
        if all_reduce_group is not None:
            data_parallel_size *= all_reduce_group.size()
        # Since fp16 has smaller dynamic range than fp32/bf16, we want to avoid
        # overflow/underflow. For N data parallel workers, each worker computes
        # g_i, and they collectively reduce (g_1 + ... + g_N) / N. To avoid
        # overflow/underflow, we divide by ~sqrt(N) before/after the reduction.
        factor: int = 1
        while data_parallel_size % factor == 0 and data_parallel_size / factor > factor:
            factor *= 2
        factor = float(factor)
        return (factor, data_parallel_size / factor)

    from torch.distributed.fsdp._fully_shard._fsdp_common import (
        _get_dim0_padded_size,
        _raise_assert_with_print,
        compiled_autograd_enabled,
    )

    """
    This is a modified version of `torch.distributed.fsdp.utils.foreach_reduce`
    that supports custom gradient accumulation on promoted data types.

    The original `foreach_reduce` accumulates gradients in the original data type, which is not
    appropriate for low-precision like bf16, especially for large number of mini batches to accumulate with.
    """

    @torch.no_grad()
    def foreach_reduce(
        fsdp_params: list[FSDPParam],
        unsharded_grads: list[torch.Tensor],
        reduce_scatter_group: dist.ProcessGroup,
        reduce_scatter_stream: torch.Stream,
        orig_dtype: torch.dtype,
        reduce_dtype: Optional[torch.dtype],
        device: torch.device,
        reduce_scatter_reduce_op: Optional[
            Union[dist.ReduceOp, dist.ReduceOp.RedOpType]
        ],
        all_reduce_group: Optional[dist.ProcessGroup],  # not `None` iff HSDP
        all_reduce_stream: torch.Stream,
        all_reduce_grads: bool,
        partial_reduce_output: Optional[torch.Tensor],  # only used for HSDP
        all_reduce_hook: Optional[
            Callable[[torch.Tensor], None]
        ] = None,  # Compatible for 2.6.0
    ) -> tuple[
        torch.Tensor,
        torch.Event,
        torch.Event,
        Optional[torch.Tensor],
        Optional[torch.Event],
        Optional[torch.Tensor],
    ]:
        if torch.__version__ >= "2.8.0":
            logger.error(
                f"This version of torch={torch.__version__} is not supported for gradient accumulation on promoted data types."
            )
            raise NotImplementedError(
                f"This version of torch={torch.__version__} is not supported for gradient accumulation on promoted data types."
            )

        """
        ``unsharded_grads`` owns the references to the gradients computed by
        autograd, so clearing the list frees the gradients.
        """
        grad_dtypes = {grad.dtype for grad in unsharded_grads}
        if len(grad_dtypes) != 1:
            # Check this at runtime since it could be a real runtime error if e.g.
            # fp8 weights do not produce the correct higher precision gradients
            _raise_assert_with_print(
                f"FSDP reduce-scatter expects uniform gradient dtype but got {grad_dtypes}"
            )
        grad_dtype = unsharded_grads[0].dtype
        reduce_dtype = reduce_dtype or grad_dtype
        predivide_factor, postdivide_factor = _get_gradient_divide_factors(
            reduce_scatter_group, all_reduce_group, reduce_dtype, device.type
        )
        world_size = reduce_scatter_group.size()
        for i, (fsdp_param, unsharded_grad) in enumerate(
            zip(fsdp_params, unsharded_grads)
        ):
            if (shard_dim := fsdp_param.fsdp_placement.dim) == 0:
                continue
            assert (
                unsharded_grad.size(shard_dim) % world_size == 0
            ), f"Shard({shard_dim}) requires even sharding: {unsharded_grad.size()=} {world_size=}"
            chunks = torch.chunk(unsharded_grad, world_size, dim=shard_dim)
            unsharded_grads[i] = torch.cat(chunks, dim=0)
        padded_unsharded_sizes = tuple(
            _get_dim0_padded_size(grad.size(), world_size) for grad in unsharded_grads
        )
        reduce_scatter_input_numel = sum(s.numel() for s in padded_unsharded_sizes)
        reduce_scatter_output_numel = reduce_scatter_input_numel // world_size
        reduce_scatter_input = torch.empty(
            (reduce_scatter_input_numel,), dtype=reduce_dtype, device=device
        )
        device_handle = _get_device_handle(device.type)
        foreach_reduce_scatter_copy_in(
            unsharded_grads, reduce_scatter_input, world_size
        )
        current_stream = device_handle.current_stream()
        # Only after the copy-in finishes can we free the gradients
        unsharded_grads.clear()
        reduce_scatter_stream.wait_stream(current_stream)
        all_reduce_input = None
        all_reduce_event = None
        with device_handle.stream(reduce_scatter_stream):
            reduce_output = reduce_scatter_input.new_empty(
                (reduce_scatter_output_numel,)
            )
            _div_if_needed(reduce_scatter_input, predivide_factor)
            if reduce_scatter_reduce_op is None:
                if predivide_factor is None:
                    reduce_scatter_reduce_op = ReduceOp.AVG
                else:
                    reduce_scatter_reduce_op = ReduceOp.SUM
            dist.reduce_scatter_tensor(
                output=reduce_output,
                input=reduce_scatter_input,
                group=reduce_scatter_group,
                op=reduce_scatter_reduce_op,
            )
            reduce_scatter_event = reduce_scatter_stream.record_event()
            post_reduce_stream = reduce_scatter_stream
            if all_reduce_group is not None:  # HSDP
                # Accumulations must run in the reduce-scatter stream
                if not all_reduce_grads:
                    if partial_reduce_output is not None:
                        partial_reduce_output += reduce_output
                    else:
                        partial_reduce_output = reduce_output
                    return (
                        reduce_scatter_input,
                        reduce_scatter_event,
                        post_reduce_stream.record_event(),
                        all_reduce_input,
                        all_reduce_event,
                        partial_reduce_output,
                    )
                if partial_reduce_output is not None:
                    reduce_output += partial_reduce_output
                post_reduce_stream = all_reduce_stream
                all_reduce_stream.wait_stream(reduce_scatter_stream)
                with device_handle.stream(all_reduce_stream):
                    dist.all_reduce(
                        reduce_output,
                        group=all_reduce_group,
                        op=ReduceOp.AVG if predivide_factor is None else ReduceOp.SUM,
                    )
                    all_reduce_input = reduce_output
                    all_reduce_event = all_reduce_stream.record_event()
        # -- END: ops in reduce_scatter stream

        if all_reduce_hook is not None:
            # Execute user-specified all reduce hook.
            # If native HSDP is used, this is executed after the HSDP all reduce.
            # If 1-d FSDP is used, this is executed post reduce-scatter.
            post_reduce_stream = all_reduce_stream
            all_reduce_stream.wait_stream(reduce_scatter_stream)
            with device_handle.stream(all_reduce_stream):
                all_reduce_hook(reduce_output)
        # -- END: ops post reduce_scatter

        with device_handle.stream(post_reduce_stream):
            _div_if_needed(reduce_output, postdivide_factor)
            # reduce_output = _to_dtype_if_needed(reduce_output, orig_dtype)
            # View out and accumulate sharded gradients
            flat_grad_offset = 0  # [0, reduce_scatter_output_numel - 1]
            for padded_unsharded_size, fsdp_param in zip(
                padded_unsharded_sizes, fsdp_params
            ):
                # Assume even sharding for Shard(i), i > 0; otherwise would require
                # copy-out for contiguous strides
                new_sharded_grad = torch.as_strided(
                    reduce_output,
                    size=fsdp_param.sharded_size,
                    stride=fsdp_param.contiguous_sharded_stride,
                    storage_offset=flat_grad_offset,
                )
                to_accumulate_grad = (
                    hasattr(fsdp_param.sharded_param, "cosmos_grad")
                    and fsdp_param.sharded_param.cosmos_grad is not None
                )
                if fsdp_param.offload_to_cpu:
                    # Only overlap the D2H copy (copying to pinned memory) if not
                    # accumulating gradients since the CPU add kernel depends on
                    # the copy result and we cannot run the add as a callback
                    non_blocking = fsdp_param.pin_memory and not to_accumulate_grad
                    # Since the GPU sharded gradient is allocated in the RS stream,
                    # we can free it here by not keeping a ref without waiting for
                    # the D2H copy since future RS-stream ops run after the copy
                    new_sharded_grad = new_sharded_grad.to(
                        torch.device("cpu"), non_blocking=non_blocking
                    )
                    if non_blocking:
                        # Record an event on which to block the CPU thread to
                        # ensure that the D2H copy finishes before the optimizer
                        fsdp_param.grad_offload_event = (
                            reduce_scatter_stream.record_event()
                        )
                if to_accumulate_grad:
                    assert isinstance(fsdp_param.sharded_param.cosmos_grad, DTensor)
                    fsdp_param.sharded_param.cosmos_grad._local_tensor += (
                        new_sharded_grad
                    )
                else:
                    new_sharded_dtensor_grad = fsdp_param.to_sharded_dtensor(
                        new_sharded_grad
                    )
                    fsdp_param.sharded_param.cosmos_grad = new_sharded_dtensor_grad
                if not compiled_autograd_enabled():
                    for hook in (
                        getattr(
                            fsdp_param.sharded_param, "_post_accumulate_grad_hooks", {}
                        )
                        or {}
                    ).values():
                        hook(fsdp_param.sharded_param)
                padded_sharded_numel = padded_unsharded_size.numel() // world_size
                flat_grad_offset += padded_sharded_numel
            post_reduce_event = post_reduce_stream.record_event()
        # The RS output is allocated in the RS stream and used in the default
        # stream (for optimizer). To ensure its memory is not reused for later
        # RSs, we do not need extra synchronization since the sharded parameters
        # hold refs through the end of backward.
        return (
            reduce_scatter_input,
            reduce_scatter_event,
            post_reduce_event,
            all_reduce_input,
            all_reduce_event,
            None,
        )

    # Patch the `foreach_reduce` function in `torch.distributed.fsdp._fully_shard._fsdp_collectives`
    import torch.distributed.fsdp._fully_shard._fsdp_collectives as fsdp_collectives

    fsdp_collectives.foreach_reduce = foreach_reduce
    import torch.distributed.fsdp._fully_shard._fsdp_param_group as pg

    pg.foreach_reduce = foreach_reduce



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from torch.utils.data import Sampler
from typing import Iterator


class SkippingSampler(Sampler[int]):
    """
    One-shot wrapper around an index-level Sampler that skips `skip_samples`
    indices once, then behaves like the base sampler thereafter.
    """

    def __init__(self, base_sampler: Sampler[int], skip_samples: int = 0):
        self.base = base_sampler
        self._initial_skip = max(0, int(skip_samples))
        self._remaining_skip = self._initial_skip
        self._used_once = False

    def __iter__(self) -> Iterator[int]:
        it = iter(self.base)
        if self._remaining_skip > 0:
            for _ in range(self._remaining_skip):
                try:
                    next(it)
                except StopIteration:
                    self._remaining_skip = 0
                    self._used_once = True
                    return iter(())
            self._remaining_skip = 0
        self._used_once = True
        return it

    def __len__(self) -> int:
        base_len = len(self.base)
        if not self._used_once:
            return max(0, base_len - self._initial_skip)
        return base_len

    def set_epoch(self, epoch: int):
        self.base.set_epoch(epoch)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Any, List


REGISTERED_ALGOs = {}


class RuleBasedAlgo(ABC):
    @abstractmethod
    def compute_reward(self, to_be_evaluated: Any, reference: Any) -> float:
        return 0.0

    @abstractmethod
    def compute_advantage(self, rewards: List[float]) -> List[float]:
        return rewards

    @abstractmethod
    def ready(self) -> bool:
        return False


def _register_rule_based_algo(name: str, algo: RuleBasedAlgo):
    REGISTERED_ALGOs[name] = algo



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Any, List, Dict, Type, Union
from transformers import AutoTokenizer
from cosmos_rl.policy.config import Config


class DataPacker(ABC):
    _MODEL_TO_DEFAULT_DATA_PACKER_REGISTRY: Dict[str, Type["DataPacker"]] = {}

    """
    This is where dataset item is transformed into the format required by the rollout engine (e.g. vllm)
    for example:
        - `str` is needed for language model
        - {
            "prompt": prompt,
            "multi_modal_data": {"video": ...},
          } for multi-modal model
    """

    @classmethod
    def register(
        cls,
        model_types: Union[str, List[str]],
        default_data_packer_cls: Type["DataPacker"],
        *,
        allow_override: bool = False,
    ):
        if isinstance(model_types, str):
            model_types = [model_types]
        else:
            model_types = list(model_types)

        for model_type in model_types:
            if (
                not allow_override
                and model_type in DataPacker._MODEL_TO_DEFAULT_DATA_PACKER_REGISTRY
                and DataPacker._MODEL_TO_DEFAULT_DATA_PACKER_REGISTRY[model_type]
                != default_data_packer_cls
            ):
                raise ValueError(f"DataPacker for {model_type} is already registered")
            DataPacker._MODEL_TO_DEFAULT_DATA_PACKER_REGISTRY[model_type] = (
                default_data_packer_cls
            )

    @classmethod
    def get_default_data_packer(cls, model_type: str) -> Type["DataPacker"]:
        if model_type not in DataPacker._MODEL_TO_DEFAULT_DATA_PACKER_REGISTRY:
            raise ValueError(f"DataPacker for {model_type} is not registered")
        return DataPacker._MODEL_TO_DEFAULT_DATA_PACKER_REGISTRY[model_type]()

    def setup(self, config: Config, tokenizer: AutoTokenizer, *args, **kwargs):
        """
        Called by launcher after being mounted
        """
        assert config is not None, "config should be set"
        assert tokenizer is not None, "tokenizer should be set"
        self.config = config
        self.tokenizer = tokenizer

    @abstractmethod
    def get_rollout_input(self, item: Any) -> Any:
        """
        Convert sample to data format required by the rollout engine (e.g. vllm)
        """
        raise NotImplementedError("This method should be implemented by the subclass")

    def rollout_collate_fn(self, items: List[Any]) -> List[Any]:
        """
        Collate the rollout inputs into a mini-batch for rollout engine
        """
        return items

    @abstractmethod
    def get_policy_input(
        self,
        sample: Any,
        rollout_output: str,
        n_ignore_prefix_tokens: int = 0,
    ) -> Any:
        """
        Stage for processing samples & rollout output before collating them into a mini-batch
        """
        raise NotImplementedError("This method should be implemented by the subclass")

    @abstractmethod
    def policy_compute_max_len(self, processed_samples: List[Any]) -> int:
        """
        Compute the maximum sequence length of the mini-batch
        """
        raise NotImplementedError("This method should be implemented by the subclass")

    @abstractmethod
    def policy_collate_fn(
        self,
        processed_samples: List[Any],
        computed_max_len: int,
    ) -> Dict[str, Any]:
        """
        Collate the processed samples into a mini-batch,
        the output will be fed into the policy model for training
        """
        raise NotImplementedError("This method should be implemented by the subclass")

    def sft_process_sample(self, sample: Any) -> Any:
        """
        Process the sample into the format required by the SFT model
        """
        raise NotImplementedError(
            "This method should be implemented by the subclass for SFT training"
        )

    def sft_compute_max_len(self, processed_samples: List[Any]) -> int:
        """
        Compute the maximum sequence length of the processed samples
        """
        raise NotImplementedError(
            "This method should be implemented by the subclass for SFT training"
        )

    def sft_collate_fn(
        self,
        sub_batch: List[Dict[str, Any]],
        computed_max_len: int,
        pad_token_id: int,
        ignore_label_id: int,
    ) -> Dict[str, Any]:
        """
        Collate the mini-batch into the format required by the policy model
        """
        raise NotImplementedError(
            "This method should be implemented by the subclass for SFT training"
        )



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from functools import cached_property
from typing import Any, Callable, List, Optional, Tuple, Union

import torch
import torch.nn as nn
from safetensors import safe_open
from torch.nn.modules.module import _IncompatibleKeys
from transformers import AutoConfig

try:
    from torch.distributed.tensor import DTensor
except ImportError:
    print("torch.distributed.tensor is not available. DeepSeek model will not work.")


from cosmos_rl.dispatcher.data.packer.deepseek_data_packer import DeepSeek_DataPacker
from cosmos_rl.policy.config import Config as CosmosConfig
from cosmos_rl.policy.kernel.moe import moe
from cosmos_rl.policy.model.base import BaseModel, ModelRegistry
from cosmos_rl.policy.model.deepseek_v3 import deepseekv3_mapped
from cosmos_rl.policy.model.deepseek_v3.checkpoint_planner import RenameLoadPlanner
from cosmos_rl.policy.model.deepseek_v3.weight_mapper import (
    DeepseekV3MoEWeightMapper,
    convert_weight_from_hf,
    weight_dequant,
)
from cosmos_rl.utils.logging import logger
from cosmos_rl.utils.parallelism import ParallelDims
from cosmos_rl.utils.util import clear_weight_name, resolve_model_path, retry

DCP_CHECKPOINT_PATH_PREFIX = "/root/.cache"
DCP_CHECKPOINT_PATH_SUFFIX = "dcp"


@ModelRegistry.register(
    DeepseekV3MoEWeightMapper, default_data_packer_cls=DeepSeek_DataPacker
)
class DeepseekV3MoEModel(BaseModel):
    @staticmethod
    def supported_model_types():
        return ["deepseek_v3"]

    @classmethod
    def from_pretrained(
        cls,
        hf_config: AutoConfig,
        model_name_or_path: str,
        max_position_embeddings: Optional[int] = None,
    ) -> "DeepseekV3MoEModel":
        """
        Initialize a DeepseekV3MoE model from a pretrained model.

        Args:
            model_name_or_path (str): Model name or path to the pretrained model.

        Returns:
            DeepseekV3MoE: DeepseekV3MoE model.

        """
        if hf_config.model_type not in cls.supported_model_types():
            raise ValueError(f"Unsupported model type: {hf_config.model_type}")

        assert (
            "deepseek-v3" in model_name_or_path.lower()
        ), f"Unsupported model {model_name_or_path}"

        if hf_config.num_hidden_layers == 4:
            deepseek_config = deepseekv3_mapped.DeepseekConfig(n_layers=4)
        elif hf_config.num_hidden_layers == 61:
            deepseek_config = deepseekv3_mapped.DeepseekConfig()
        else:
            raise ValueError(
                f"Only 4 or 61 layer models supported at the moment. Got hf_config.llm_config.num_hidden_layers={hf_config.llm_config.num_hidden_layers}"
            )

        model = DeepseekV3MoEModel(deepseek_config, hf_config=hf_config)

        logger.info("Initializing the model")
        with torch.no_grad():
            model.init_weights()
        logger.info("Model initialized")
        return model

    def __init__(
        self,
        model_config: deepseekv3_mapped.DeepseekConfig,
        hf_config: AutoConfig,
    ):
        super().__init__(hf_config=hf_config)
        self.config = model_config

        orig_precision = torch.get_default_dtype()
        precision = getattr(torch, model_config.dtype)
        torch.set_default_dtype(precision)
        logger.info(f"Setting torch default dtype from {orig_precision} to {precision}")

        self.build_model(model_config)

        torch.set_default_dtype(
            orig_precision
        )  # Reset the default dtype to the original value
        logger.info(f"Reset torch default dtype to {orig_precision}")

    def build_model(self, model_config: deepseekv3_mapped.DeepseekConfig):
        # Create reasoning model
        self.model = deepseekv3_mapped.Transformer(args=model_config)

    def init_weights(
        self,
        buffer_device: Optional[torch.device] = None,
    ):
        self.apply(_init_weights)

    def forward(
        self,
        input_ids: torch.Tensor = None,
        position_ids: torch.Tensor = None,
        initial_aux_loss: Optional[torch.Tensor] = None,
        **data_batch: Optional[torch.Tensor],
    ) -> torch.Tensor:
        logger.debug(
            f"[model] input ids shape: {input_ids.shape}, position_ids: {position_ids.shape}"
        )

        logits, aux_loss = self.model(
            tokens=input_ids,
            position_ids=position_ids,
            padding_mask=data_batch.get("padding_mask", None),
        )

        if self.config.aux_loss_coeff > 0:
            if initial_aux_loss is not None and aux_loss is not None:
                final_aux_loss = initial_aux_loss + aux_loss
                return logits, final_aux_loss
            elif initial_aux_loss is not None:
                final_aux_loss = initial_aux_loss
                return logits, final_aux_loss
            else:
                return logits
        else:
            assert (
                initial_aux_loss is None
            ), "initial_aux_loss must be None when aux_loss_coeff = 0"
            assert aux_loss is None, "aux_loss must be None when aux_loss_coeff = 0"
            return logits

    @property
    def parallelize_fn(self):
        from cosmos_rl.policy.model.deepseek_v3.parallelize import parallelize_model

        return parallelize_model, self

    def post_to_empty_hook(self, cosmos_config: CosmosConfig):
        return

    def separate_model_parts(self) -> List[nn.Module]:
        return [self]

    def get_position_ids(self, **kwargs) -> Tuple[torch.Tensor, torch.Tensor, int]:
        seq_dim_idx = 1
        inputs = kwargs["input_ids"]
        position_ids = (
            torch.arange(inputs.size(-1), dtype=torch.long, device=inputs.device)
            .unsqueeze(0)
            .expand_as(inputs)
        )
        return position_ids, inputs, seq_dim_idx

    def apply_pipeline_split(self, pp_rank, pp_size):
        raise NotImplementedError

    @cached_property
    def _get_nparams_and_flops_fn(self) -> Callable[[int], tuple[int, int]]:
        nparams = sum(p.numel() for p in self.parameters())
        nparams_embedding = sum(
            sum(p.numel() for p in m.parameters())
            for m in self.children()
            if isinstance(m, nn.Embedding)
        )

        # Reasoning behind the factor of 12 for the self-attention part of the formula:
        # 1. each self-attention has 2 matmul in the forward and 4 in the backward (6)
        # 2. the flash attention does 1 more matmul recomputation in the backward
        #    but recomputation should not be counted in calculating MFU           (+0)
        # 3. each matmul performs 1 multiplication and 1 addition                 (*2)
        # 4. we follow the convention and do not account for sparsity in causal attention
        layers, heads, head_dim = (
            self.config.n_layers,
            self.config.n_heads,
            self.config.dim // self.config.n_heads,
        )
        return lambda seq_len: (
            nparams,
            6 * (nparams - nparams_embedding)
            + 12 * layers * heads * head_dim * seq_len,
        )

    def get_nparams_and_flops(self, seq_len: int) -> tuple[int, int]:
        return self._get_nparams_and_flops_fn(seq_len)

    def load_hf_weights(
        self,
        model_name_or_path: str,
        parallel_dims: ParallelDims,
        device: torch.device,
        revision: Optional[str] = None,
    ):
        dcp_checkpoint_path = os.path.join(
            DCP_CHECKPOINT_PATH_PREFIX,
            model_name_or_path.split("/")[-1].lower(),
            DCP_CHECKPOINT_PATH_SUFFIX,
        )
        # if it is a huggingface model and no checkpoint exists, we need to load the weights from the safetensors files
        if len(model_name_or_path.split("/")) == 2 and (
            not os.path.exists(dcp_checkpoint_path)
            or len(os.listdir(dcp_checkpoint_path)) == 0
        ):
            # The checkpoint loading assumes bf16 checkpoints. The default deepseekv3 checkpoint is in fp8. We needs to convert the checkpoints from fp8 to bf16 first.
            # Can follow the instructions here: https://github.com/NVIDIA-NeMo/RL/blob/main/docs/guides/deepseek.md
            # Load all safetensors from `model_path`
            model_type = retry(AutoConfig.from_pretrained)(
                model_name_or_path, trust_remote_code=True
            ).model_type
            model_path = resolve_model_path(model_name_or_path, revision)
            safetensors_files = [
                f for f in os.listdir(model_path) if f.endswith(".safetensors")
            ]

            self_state_dict = self.state_dict()
            self_state_dict = {
                clear_weight_name(k): v for k, v in self_state_dict.items()
            }

            lm_head_weight_key = "model.lm_head.weight"
            embed_tokens_weight_key = "model.model.embed_tokens.weight"
            weights_of_ckpt_names = set()
            reserved = {}
            scale_inv_paths = {}

            for f in safetensors_files:
                ckpt = retry(safe_open)(
                    os.path.join(model_path, f), framework="pt", device=str(device)
                )
                keys = ckpt.keys()
                for name in keys:
                    if name.endswith("weight_scale_inv"):
                        scale_inv_paths[name] = os.path.join(model_path, f)

            for f in safetensors_files:
                logger.info(f"Loading safetensors: {f}")
                weights_of_ckpt = {}
                ckpt = retry(safe_open)(
                    os.path.join(model_path, f), framework="pt", device=str(device)
                )
                keys = ckpt.keys()
                for name in keys:
                    ckpt_tensor = ckpt.get_tensor(name)
                    weights_of_ckpt[name] = ckpt_tensor
                    weights_of_ckpt_names.add(name)
                    if name == embed_tokens_weight_key:
                        reserved[name] = ckpt_tensor

                for name in weights_of_ckpt.keys():
                    tensor = weights_of_ckpt[name]
                    if name.endswith("weight_scale_inv") or "layers.61" in name:
                        # Skip since this weight is used for dequantization
                        continue

                    if (
                        "down_proj" in name
                        or "up_proj" in name
                        or "gate_proj" in name
                        or "self_attn.kv_a_proj_with_mqa" in name
                        or "self_attn.kv_b_proj" in name
                        or "self_attn.o_proj" in name
                        or "self_attn.q_a_proj" in name
                        or "self_attn.q_b_proj" in name
                    ) and "weight" in name:
                        inv_name = name + "_scale_inv"
                        inv_tensor = retry(safe_open)(
                            scale_inv_paths[inv_name],
                            framework="pt",
                            device=str(device),
                        ).get_tensor(inv_name)
                        tensor = weight_dequant(tensor, inv_tensor)

                    dest_name, shared_weight, expert_id = convert_weight_from_hf(
                        tensor,
                        name,
                        model_type,
                        parallel_dims,
                        n_experts=self.config.n_routed_experts,
                    )

                    if dest_name is None:
                        # This is due to the expert parallelism grouping
                        continue

                    if dest_name not in self_state_dict and parallel_dims.pp_enabled:
                        logger.info(
                            f"Weight `{dest_name}` is discarded, maybe due to pipeline parallelism or expert parallelism grouping. Skipping this weight checking"
                        )
                        continue

                    target_tensor = self_state_dict[dest_name]
                    if isinstance(target_tensor, torch.distributed.tensor.DTensor):
                        target_tensor = target_tensor.to_local()
                    # Write to the correct expert of the target tensor
                    if expert_id is not None:
                        # Convert expert_id to local_expert_id
                        n_local_experts = (
                            self.config.n_routed_experts
                            // parallel_dims.tp
                            // parallel_dims.dp_shard
                        )

                        expert_id = expert_id % n_local_experts
                        target_tensor = target_tensor[expert_id]

                    assert (
                        target_tensor.shape == shared_weight.shape
                    ), f"Shape mismatch: {target_tensor.shape} != {shared_weight.shape} for {dest_name}"
                    with torch.no_grad():
                        target_tensor.data.copy_(shared_weight)
                torch.distributed.barrier()
                logger.info(f"Loaded safetensors: {f} successfully.")

            if (
                lm_head_weight_key not in weights_of_ckpt_names
                and embed_tokens_weight_key in weights_of_ckpt_names
            ):
                # tied with embed_tokens.weight
                name = lm_head_weight_key
                assert embed_tokens_weight_key in reserved
                tensor = reserved[embed_tokens_weight_key]
                dest_name, shared_weight = convert_weight_from_hf(
                    tensor,
                    name,
                    model_type,
                    parallel_dims,
                    n_experts=self.config.n_routed_experts,
                )
                if dest_name in self_state_dict:
                    target_tensor = self_state_dict[dest_name]
                    is_dist_tensor = isinstance(
                        target_tensor, torch.distributed.tensor.DTensor
                    )
                    local_view = (
                        target_tensor.to_local() if is_dist_tensor else target_tensor
                    )
                    assert (
                        local_view.shape == shared_weight.shape
                    ), f"Shape mismatch: {local_view.shape} != {shared_weight.shape} for {dest_name}"
                    with torch.no_grad():
                        local_view.data.copy_(shared_weight)

            logger.info(f"Dumping the tensors to DCP folder {dcp_checkpoint_path}")
            os.makedirs(dcp_checkpoint_path, exist_ok=True)
            fs_storage_writer = torch.distributed.checkpoint.FileSystemWriter(
                dcp_checkpoint_path
            )
            torch.distributed.checkpoint.save(
                state_dict=self_state_dict,
                storage_writer=fs_storage_writer,
            )
        else:
            logger.info("Loading from distributed checkpoints...")
            model_name_or_path = model_name_or_path.rstrip("/")
            if model_name_or_path.endswith("_hf"):
                model_name_or_path_dcp = model_name_or_path[:-3]
                logger.info(
                    f"Found model path with _hf prefix ({model_name_or_path}. Looking for non-hf checkpoint at: {model_name_or_path_dcp}"
                )
                model_name_or_path = model_name_or_path_dcp
            elif len(model_name_or_path.split("/")) == 2:
                model_name_or_path = dcp_checkpoint_path

            # Transformer engine adds this extra state which we dont have in the saved checkpoint.
            mapped_state_dict = {
                k: v
                for k, v in self.state_dict().items()
                if not k.endswith("_extra_state")
            }

            logger.info("Creating storage reader...")
            fs_storage_reader = torch.distributed.checkpoint.FileSystemReader(
                model_name_or_path
            )
            logger.info("Loading checkpoint ...")
            torch.distributed.checkpoint.load(
                state_dict=mapped_state_dict,
                storage_reader=fs_storage_reader,
                planner=RenameLoadPlanner(allow_partial_load=False),
            )
            logger.info("Refreshing model.")
            self.load_state_dict(self.state_dict())
            logger.info("Checkpoint loaded.")

    def load_state_dict(
        self, state_dict: dict[str, Any], strict: bool = True, assign: bool = False
    ):
        """
        Ignore the missing keys with substrings matching `substring_to_ignore` (e.g., "_extra_state" keys imposed by
        TransformerEngine for FP8).
        """
        actual_missing_keys, unexpected_keys = super().load_state_dict(
            state_dict, strict=False, assign=assign
        )
        if strict:
            if len(actual_missing_keys) > 0 or len(unexpected_keys) > 0:
                raise ValueError(
                    f"Missing keys: {actual_missing_keys}\n\nUnexpected keys: {unexpected_keys}"
                )
        return _IncompatibleKeys(actual_missing_keys, unexpected_keys)

    @cached_property
    def weight_sync_transforms(
        self,
    ) -> List[Tuple[str, Union[torch.Tensor, Callable]]]:
        # 1. get all parameters, but not buffers
        transforms = {}

        for local_name, param in self.named_parameters():
            hf_name = self.weight_mapper.policy_map_local_key_to_hf_key(
                clear_weight_name(local_name)
            )

            is_dist_tensor = isinstance(param, torch.distributed.tensor.DTensor)
            transform_or_view = param.to_local() if is_dist_tensor else param

            assert (
                hf_name not in transforms
            ), f"Duplicate key found in transforms: {hf_name}"
            transforms[hf_name] = transform_or_view

        return sorted(transforms.items())


def _init_weights(module):
    std = 0.02

    def to_local(tensor):
        if isinstance(tensor, DTensor):
            return tensor.to_local()
        else:
            return tensor

    if isinstance(module, torch.nn.Linear):
        to_local(module.weight).normal_(mean=0.0, std=std)
        if module.bias is not None:
            to_local(module.bias).zero_()
    elif isinstance(module, torch.nn.Embedding):
        to_local(module.weight).normal_(mean=0.0, std=std)
        if module.padding_idx is not None:
            to_local(module.weight)[module.padding_idx].zero_()
    elif isinstance(module, moe.Gate):
        to_local(module.weight).normal_(mean=0.0, std=std)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import types
import torch

from typing import Any, Tuple, Callable
from functools import partial

from torch.distributed.device_mesh import DeviceMesh
import torch.distributed as dist
import torch.nn as nn

from cosmos_rl.utils.attn_util import repeat_kv
from cosmos_rl.utils.parallelism import ParallelDims


def all_to_all_tensor(
    local_input: torch.Tensor,
    scatter_dim: int,
    gather_dim: int,
    cp_mesh: DeviceMesh,
    async_op: bool = False,
) -> torch.Tensor:
    """
    All‐to‐all via all_to_all_single.
    Splits `local_input` along scatter_dim into cp_world_size pieces,
    exchanges them, then concatenates received pieces along gather_dim.
    Returns a tensor or, if async_op=True, returns a callable that waits and returns.
    """
    group = cp_mesh.get_group()
    world_size = cp_mesh.size()
    # 1) Split into per‐rank chunks and record each chunk's shape & numel
    splits = torch.tensor_split(local_input, world_size, dim=scatter_dim)
    split_shapes = [s.shape for s in splits]
    split_nelems = [s.numel() for s in splits]

    # 2) Flatten & concat all send‐chunks into one 1‐D tensor
    flat_input = torch.cat([s.contiguous().view(-1) for s in splits], dim=0)
    flat_output = torch.empty_like(flat_input)

    # 3) Exchange
    work = dist.all_to_all_single(
        flat_output,  # output
        flat_input,  # input
        output_split_sizes=split_nelems,
        input_split_sizes=split_nelems,
        group=group,
        async_op=async_op,
    )

    # helper to rebuild & cat along gather_dim
    def _build_result():
        out_chunks = []
        offset = 0
        for shape, ne in zip(split_shapes, split_nelems):
            chunk = flat_output[offset : offset + ne].view(shape)
            out_chunks.append(chunk)
            offset += ne
        return torch.cat(out_chunks, dim=gather_dim).contiguous()

    if async_op:

        def wait_and_get():
            work.wait()
            return _build_result()

        return wait_and_get

    # synchronous
    return _build_result()


def all_gather_tensor(
    local_tensor: torch.Tensor, cp_mesh: DeviceMesh, async_op: bool = False
) -> torch.Tensor:
    """
    This function performs an all-gather communication operation on a tensor.
    It splits the input tensor into `cp_world_size` parts along the specified scatter dimension,
    and then performs an all-to-all communication operation on these parts.
    The output is a tensor of the same shape as the input, but with the specified gather dimension
    concatenated.

    Args:
        local_tensor (torch.Tensor): The input tensor to be gathered.
        cp_mesh (DeviceMesh): The device mesh to use for the all-gather communication.
        async_op (bool, optional): Whether to perform the operation asynchronously. Defaults to False.

    Returns:
        torch.Tensor: The output tensor of the same shape as the input, but with the specified gather dimension concatenated.
    """
    group = cp_mesh.get_group()
    cp_world_size = cp_mesh.size()
    output_shape = list(local_tensor.shape)
    output_shape[0] = output_shape[0] * cp_world_size
    output = torch.empty(
        output_shape, dtype=local_tensor.dtype, device=local_tensor.device
    )
    comm = dist.all_gather_into_tensor(
        output, local_tensor, group=group, async_op=async_op
    )
    if async_op:

        def wait():
            comm.wait()
            return output

        return wait
    return output


class Gather(torch.autograd.Function):
    @staticmethod
    def forward(
        ctx: Any,
        cp_mesh: DeviceMesh,
        local_tensor: torch.Tensor,
        gather_dim: int,
        grad_scaler: bool = True,
        async_op=False,
    ) -> torch.Tensor:
        # Record the autograd context.
        ctx.cp_mesh = cp_mesh
        ctx.gather_dim = gather_dim
        ctx.grad_scaler = grad_scaler
        ctx.async_op = async_op

        cp_world_size = cp_mesh.size()
        ctx.cp_world_size = cp_world_size

        cp_rank = cp_mesh.get_local_rank()
        ctx.cp_rank = cp_rank

        local_shape = local_tensor.shape
        split_size = local_shape[0]
        ctx.cp_part_size = local_shape[gather_dim]

        output = all_gather_tensor(local_tensor, cp_mesh, async_op)
        return torch.cat(output.split(split_size, dim=0), dim=gather_dim)

    @staticmethod
    def backward(ctx: Any, grad_output: torch.Tensor) -> Any:
        if ctx.grad_scaler:
            grad_output = grad_output * ctx.cp_world_size
        return (
            None,
            grad_output.split(ctx.cp_part_size, dim=ctx.gather_dim)[
                ctx.cp_rank
            ].contiguous(),
            None,
            None,
            None,
            None,
        )


class SeqAllToAll(torch.autograd.Function):
    """
    This class implements the all-to-all communication operation for a sequence of tensors.
    """

    @staticmethod
    def forward(
        ctx: Any,
        cp_mesh: DeviceMesh,
        local_input: torch.Tensor,
        scatter_dim: int,
        gather_dim: int,
        async_op: bool = False,
    ) -> torch.Tensor:
        ctx.cp_mesh = cp_mesh
        ctx.scatter_dim = scatter_dim
        ctx.gather_dim = gather_dim
        ctx.async_op = async_op
        return all_to_all_tensor(
            local_input, scatter_dim, gather_dim, cp_mesh, async_op
        )

    @staticmethod
    def backward(
        ctx: Any, *grad_output: torch.Tensor
    ) -> Tuple[None, torch.Tensor, None, None]:
        input_t = (
            torch.cat(grad_output[1:], dim=ctx.gather_dim).contiguous()
            if ctx.async_op
            else grad_output[0]
        )
        return (
            None,
            all_to_all_tensor(
                input_t, ctx.gather_dim, ctx.scatter_dim, ctx.cp_mesh, False
            ),
            None,
            None,
            None,
            None,
        )


def slice_input_tensor(
    data: torch.Tensor, dim: int, cp_mesh: DeviceMesh
) -> torch.Tensor:
    # We must use local rank, not get_rank()
    cp_world_size, cp_rank = cp_mesh.size(), cp_mesh.get_local_rank()
    partial_size = data.size(dim) // cp_world_size
    slc = [slice(None)] * len(data.shape)
    slc[dim] = slice(cp_rank * partial_size, (cp_rank + 1) * partial_size)
    return data[slc].contiguous()


def slice_inputs_for_ulysses(
    input_tensors: list[torch.Tensor | None],
    cp_mesh: DeviceMesh,
    seq_dims: list[int] | None = None,
) -> list[torch.Tensor]:
    """
    The input tensors are already padded by cosmos-rl datapacker.
    Args:
        input_tensors: Input tensors with shape of [bsz, seqlen]
        cp_mesh: ulysses sequence parallelism size

    Returns:
        list[torch.Tensor]: Input tensors for current CP rank.
        torch.Tensor: position_ids for current rank
    """
    if seq_dims is None:
        seq_dims = [1] * len(input_tensors)
    return [
        slice_input_tensor(t, dim=dim, cp_mesh=cp_mesh) if t is not None else None
        for t, dim in zip(input_tensors, seq_dims)
    ]


def gather_outputs_for_ulysses(
    output: torch.Tensor,
    gather_dim: int,
    cp_mesh: DeviceMesh,
    grad_scaler: bool = True,
) -> torch.Tensor:
    return Gather.apply(cp_mesh, output, gather_dim, grad_scaler)


def gather_seq_scatter_heads(
    x: torch.Tensor,
    seq_dim: int,
    head_dim: int,
    cp_mesh: DeviceMesh,
) -> torch.Tensor:
    """
    A func to sync embedding input with alltoall in sequence parallel.
    gather sequence dimension and scatter head dim:
    For example, when seq_dim is 1, head_dim is 2, the transformation is:
    [bsz, seq/n, h, ...] -> [bsz, seq, h/n, ...]
    Args:
        x: shape of [bsz, seq/n, h, ...]
        seq_dim: the dimension to gather
        head_dim: the dimension to scatter
        cp_mesh: ulysses sequence parallelism size
    Returns:
        torch.Tensor: shape of gathered and scattered tensor
    """
    x = SeqAllToAll.apply(cp_mesh, x, head_dim, seq_dim)
    return x


def gather_heads_scatter_seq(
    x: torch.Tensor, head_dim: int, seq_dim: int, cp_mesh: DeviceMesh
) -> torch.Tensor:
    """
    A func to sync attention result with alltoall in sequence parallel.
    gather head dimension and scatter seq dim:
    For example, when seq_dim is 1, head_dim is 2, the transformation is:
    [bsz, seq, h/n, ...] -> [bsz, seq/n, h, ...]

    Args:
        x (torch.Tensor): shape of [bsz, seq, h/n, ...]
        head_dim (int): the dimension to gather
        seq_dim (int): the dimension to scatter
        cp_mesh (DeviceMesh): ulysses sequence parallelism size

    Returns:
        torch.Tensor: shape of [bsz, seq/n, h, ...]
    """
    return SeqAllToAll.apply(cp_mesh, x, seq_dim, head_dim, False)


def ulysses_wrapper_of_attn_func(
    query_states: torch.Tensor,
    key_states: torch.Tensor,
    value_states: torch.Tensor,
    cp_mesh: DeviceMesh,
    original_attn_func: Callable,
    *args,
    **kwargs,
):
    """Insert all-to-all before and after flash attention.
    DeepSpeed-Ulysses: https://arxiv.org/pdf/2309.14509

    Args:
        query_states (torch.Tensor): [batch_size, seqlen/cp_size, nheads, head_dim]
        key_states (torch.Tensor): [batch_size, seqlen/cp_size, nheads_k, head_dim]
        value_states (torch.Tensor): [batch_size, seqlen/cp_size, nheads_k, head_dim]
        cp_mesh (DeviceMesh): ulysses sequence parallelism device mesh
        original_attn_func: the original attention function

    Returns:
        torch.Tensor: [batch_size, seqlen/cp_size, nheads, head_dim]
    """
    cp_world_size = cp_mesh.size()
    assert cp_world_size > 1, "CP world size must be greater than 1"

    # AlltoAll for Ulysses
    # NOTE: repeat kv heads to be divided by sequence parallel. Instead of repeating nheads_q//nheads_k,
    # we choose to repeat cp_size//nheads_k, since flash_attention supports MQA/GQA.
    # For example:
    # - nheads_k=4, sp=8, repeats=2
    # - nheads_k=8, sp=8, repeats=1
    # - nheads_k=16, sp=8, repeats=1
    repeats = max(cp_world_size // key_states.size(2), 1)
    key_states = repeat_kv(key_states, repeats)
    value_states = repeat_kv(value_states, repeats)

    # (bsz, seq_len/n, n_head, head_dim) -> (bsz, seq_len, n_head/n, head_dim)
    query_states = gather_seq_scatter_heads(
        query_states, seq_dim=1, head_dim=2, cp_mesh=cp_mesh
    )
    key_states = gather_seq_scatter_heads(
        key_states, seq_dim=1, head_dim=2, cp_mesh=cp_mesh
    )
    value_states = gather_seq_scatter_heads(
        value_states, seq_dim=1, head_dim=2, cp_mesh=cp_mesh
    )

    # (bsz, seq_len, n_head/n, head_dim)
    attn_output = original_attn_func(
        query_states, key_states, value_states, *args, **kwargs
    )

    # AlltoAll for Ulysses
    # (bsz, seq_len, n_head/n, head_dim) -> (bsz, seq_len/n, n_head, head_dim)
    attn_output = gather_heads_scatter_seq(
        attn_output, seq_dim=1, head_dim=2, cp_mesh=cp_mesh
    )

    return attn_output


def ulysses_wrapper_of_attn_func_varlen(
    query_states: torch.Tensor,
    key_states: torch.Tensor,
    value_states: torch.Tensor,
    cu_seqlens_q,
    cu_seqlens_k,
    max_seqlen_q,
    max_seqlen_k,
    cp_mesh: DeviceMesh,
    original_attn_func: Callable,
    *args,
    **kwargs,
):
    """Insert all-to-all before and after flash attention with variable length.
    DeepSpeed-Ulysses: https://arxiv.org/pdf/2309.14509

    Args:
        query_states (torch.Tensor): [batch_size, seqlen/cp_size, nheads, head_dim]
        key_states (torch.Tensor): [batch_size, seqlen/cp_size, nheads_k, head_dim]
        value_states (torch.Tensor): [batch_size, seqlen/cp_size, nheads_k, head_dim]
        cu_seqlens_q (torch.Tensor): cumulative sequence lengths for query
        cu_seqlens_k (torch.Tensor): cumulative sequence lengths for key
        max_seqlen_q (int): maximum sequence length for query
        max_seqlen_k (int): maximum sequence length for key
        cp_mesh (DeviceMesh): ulysses sequence parallelism device mesh
        original_attn_func: the original attention function

    Returns:
        torch.Tensor: [batch_size, seqlen/cp_size, nheads, head_dim]
    """
    cp_world_size = cp_mesh.size()
    assert cp_world_size > 1, "CP world size must be greater than 1"

    # AlltoAll for Ulysses
    # NOTE: repeat kv heads to be divided by sequence parallel. Instead of repeating nheads_q//nheads_k,
    # we choose to repeat cp_size//nheads_k, since flash_attention supports MQA/GQA.
    # For example:
    # - nheads_k=4, sp=8, repeats=2
    # - nheads_k=8, sp=8, repeats=1
    # - nheads_k=16, sp=8, repeats=1
    key_states = key_states.unsqueeze(0)  # [1, seqlen/cp_size, nheads_k, head_dim]
    value_states = value_states.unsqueeze(0)  # [1, seqlen/cp_size, nheads_k, head_dim]
    query_states = query_states.unsqueeze(0)  # [1, seqlen/cp_size, nheads, head_dim]
    repeats = max(cp_world_size // key_states.size(2), 1)
    key_states = repeat_kv(key_states, repeats)
    value_states = repeat_kv(value_states, repeats)
    # (1, seq_len/n, n_head, head_dim) -> (1, seq_len, n_head/n, head_dim)
    query_states = gather_seq_scatter_heads(
        query_states, seq_dim=1, head_dim=2, cp_mesh=cp_mesh
    )
    key_states = gather_seq_scatter_heads(
        key_states, seq_dim=1, head_dim=2, cp_mesh=cp_mesh
    )
    value_states = gather_seq_scatter_heads(
        value_states, seq_dim=1, head_dim=2, cp_mesh=cp_mesh
    )
    query_states = query_states.squeeze(0)
    key_states = key_states.squeeze(0)
    value_states = value_states.squeeze(0)
    attn_output = original_attn_func(
        query_states,
        key_states,
        value_states,
        cu_seqlens_q,
        cu_seqlens_k,
        max_seqlen_q,
        max_seqlen_k,
        *args,
        **kwargs,
    )
    attn_output = attn_output.unsqueeze(0)  # [1, seq_len, nheads/n, head_dim]
    # AlltoAll for Ulysses
    # (1, seq_len, n_head/n, head_dim) -> (1, seq_len/n, n_head, head_dim)
    attn_output = gather_heads_scatter_seq(
        attn_output, seq_dim=1, head_dim=2, cp_mesh=cp_mesh
    )
    attn_output = attn_output.squeeze(0)  # [1, seqlen/cp_size, nheads, head_dim]
    return attn_output


def ulysses_attn_func(
    original_attn_func: Callable,
    cp_mesh: DeviceMesh,
):
    return partial(
        ulysses_wrapper_of_attn_func,
        original_attn_func=original_attn_func,
        cp_mesh=cp_mesh,
    )


def ulysses_attn_func_varlen(
    original_attn_func: Callable,
    cp_mesh: DeviceMesh,
):
    return partial(
        ulysses_wrapper_of_attn_func_varlen,
        original_attn_func=original_attn_func,
        cp_mesh=cp_mesh,
    )


def swizzle_cp_forward(model: nn.Module, parallel_dims: ParallelDims):
    cp_mesh = parallel_dims.mesh["cp"]
    if parallel_dims.pp_enabled:
        # because `_swizzle_pp_grpo_forward` is replaced as the `forward` of model,
        # So hook wo't work in pp. We do a cp replacement, too.
        # substitute the original forward of model of last stage
        if parallel_dims.pp_coord[0] == parallel_dims.pp_coord[1] - 1:  # last stage
            origin_forward = model.forward
            full_signature = list(inspect.signature(origin_forward).parameters.keys())

            def gather_output_forward(*args, **kwargs):
                args = args[1:]
                n_args = len(args)
                if n_args > 0:
                    signature = full_signature[:n_args]
                    for key in signature:
                        if key in kwargs:
                            kwargs.pop(key)
                outputs = origin_forward(*args, **kwargs)
                return gather_outputs_for_ulysses(
                    outputs, gather_dim=1, cp_mesh=cp_mesh
                )

            model.forward = types.MethodType(gather_output_forward, model)
    else:
        # non-pp case, just use hook is perfect.
        def gather_output_hook(model, args, output):
            return gather_outputs_for_ulysses(output, gather_dim=1, cp_mesh=cp_mesh)

        model.register_forward_hook(gather_output_hook)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import torch
import torch.distributed as dist
from cosmos_rl.utils.distributed import broadcast_object_cpu
from cosmos_rl.utils.pynccl import (
    create_nccl_uid,
    create_nccl_comm,
    nccl_send,
    nccl_recv,
)
import subprocess

# This env var may lead to bug in NCCL
# Check https://github.com/NVIDIA/nccl/issues/1721
# os.environ["NCCL_CUMEM_ENABLE"] = "0"


def main():
    local_rank = int(os.environ["LOCAL_RANK"])
    global_rank = int(os.environ["RANK"])
    world_size = int(os.environ["WORLD_SIZE"])
    torch.cuda.set_device(local_rank)
    device = torch.device(f"cuda:{local_rank}")
    dist.init_process_group(backend="cuda:nccl,cpu:gloo")

    print(f"Rank {global_rank} initialized")

    nccl_uid = broadcast_object_cpu(create_nccl_uid() if global_rank == 0 else None)
    comm = create_nccl_comm(nccl_uid, global_rank, world_size)
    print(f"Rank {global_rank} created comm {comm}")

    stream = torch.cuda.Stream()

    # Test interleaved send/recv
    with torch.cuda.stream(stream):
        if global_rank in [0, 2]:
            send_tensor = (
                torch.ones([16008, 5120], dtype=torch.float16).to(device) * global_rank
            )
            nccl_send(send_tensor, 4, comm)
            nccl_send(send_tensor, 4, comm)
        elif global_rank == 4:
            recv_tensor = torch.empty([16008, 5120], dtype=torch.float16, device=device)
            nccl_recv(recv_tensor, 0, comm)
            assert torch.allclose(recv_tensor, torch.ones_like(recv_tensor) * 0)
            nccl_recv(recv_tensor, 2, comm)
            assert torch.allclose(recv_tensor, torch.ones_like(recv_tensor) * 2)
            nccl_recv(recv_tensor, 0, comm)
            assert torch.allclose(recv_tensor, torch.ones_like(recv_tensor) * 0)
            nccl_recv(recv_tensor, 2, comm)
            assert torch.allclose(recv_tensor, torch.ones_like(recv_tensor) * 2)
    torch.cuda.synchronize()
    dist.barrier()

    dist.destroy_process_group()


if __name__ == "__main__":
    if os.environ.get("RECURSIVE_ENTROPY") is None:
        n_gpu = torch.cuda.device_count()
        print(f"Running {n_gpu} processes")
        command = [
            "torchrun",
            "--nnodes",
            "1",
            "--nproc_per_node",
            str(n_gpu),
            os.path.abspath(__file__),
        ]
        env = os.environ.copy()
        env["RECURSIVE_ENTROPY"] = "1"
        subprocess.run(command, env=env)
    else:
        main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import torch
from typing import Optional, Any
from concurrent.futures import ThreadPoolExecutor

from cosmos_rl.utils.logging import logger


class DiskCache:
    """use disk cache datasets's preprocess data"""

    def __init__(self, cache_path: str):
        self.cache_path = cache_path
        os.makedirs(self.cache_path, exist_ok=True)

        self.max_concurrent_tasks = 4
        self.max_files_per_dir = 10000
        self.executor = ThreadPoolExecutor(
            max_workers=4, thread_name_prefix="disk_cache"
        )

    def __cache_ojb_path(self, idx: int) -> str:
        # avoid too many files in a single directory, limit by filesystem
        # TODO(zjx): we can reduce files count by merge data
        subdir = str(idx // self.max_files_per_dir)
        subdir_path = os.path.join(self.cache_path, subdir)
        os.makedirs(subdir_path, exist_ok=True)
        return os.path.join(subdir_path, f"{idx}.pt")

    def __save_to_disk(self, path: str, obj: Any) -> None:
        def save_obj_helper(obj, path):
            tmp_file = f"{path}.tmp"
            try:
                # Save to temporary file first
                torch.save(obj, tmp_file)
                os.rename(tmp_file, path)
            except Exception:
                # Clean up temp file if it exists
                if os.path.exists(tmp_file):
                    os.remove(tmp_file)

        # Save to temporary file first
        self.executor.submit(save_obj_helper, obj, path)

    def set(self, idx: int, obj: Any) -> None:
        if self.executor._work_queue.qsize() > self.max_concurrent_tasks:
            return

        # Create a task to run the async operation
        cachePath = self.__cache_ojb_path(idx)
        self.__save_to_disk(cachePath, obj)

    def get(self, idx: int) -> Optional[Any]:
        cachePath = self.__cache_ojb_path(idx)
        if os.path.exists(cachePath):
            try:
                return torch.load(cachePath)
            except Exception as e:
                logger.error(f"Failed to load cache file {cachePath}: {e}")
                return None
        return None

    def clear(self) -> None:
        self.executor.shutdown(wait=False)

        if os.path.exists(self.cache_path):
            shutil.rmtree(self.cache_path)



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import os
import math
import torch
from torch import nn
import torch.nn.functional as F
import torch.distributed as dist
from safetensors import safe_open
from dataclasses import dataclass, field
from typing import Tuple, List, Optional, Callable
from transformers import AutoConfig
import torch.distributed._symmetric_memory as symm_mem
from cosmos_rl.utils.util import (
    resolve_model_path,
    IdentityLayer,
    clear_weight_name,
    sync_model_vocab,
    retry,
)
from cosmos_rl.utils.logging import logger
from .weight_converter import (
    convert_weight_from_hf,
)
from cosmos_rl.utils.parallelism import ParallelDims
from cosmos_rl.policy.kernel.symm_mem_recipes import OnDeviceAllToAllV
from cosmos_rl.policy.kernel.moe.indices import generate_permute_indices
from cosmos_rl.policy.kernel.moe.grouped_gemm import group_gemm_imp
from cosmos_rl.policy.config import Config as CosmosConfig
from cosmos_rl.policy.model.base import BaseModel
from transformers.activations import ACT2FN
from functools import cached_property
from cosmos_rl.policy.kernel.norm import RMSNorm
from cosmos_rl.policy.kernel.fused import MLPActMulFunc


@dataclass
class DeepseekV3MoeArgs:
    dim: int
    intermediate_size: int
    moe_intermediate_size: int
    n_layers: int
    n_heads: int
    n_kv_heads: int
    head_dim: int
    vocab_size: int
    max_seq_len: int
    q_lora_rank: int
    kv_lora_rank: int
    v_head_dim: int
    qk_nope_head_dim: int
    qk_rope_head_dim: int
    n_routed_experts: int
    n_shared_experts: int
    routed_scaling_factor: float
    scoring_func: str
    seq_aux: bool
    topk_method: str
    n_group: int
    topk_group: int
    first_k_dense_replace: int
    moe_layer_freq: int
    biases: List[str] = field(default_factory=lambda: [])
    attention_bias: bool = False
    norm_topk_prob: bool = True
    rope_scaling: dict = None
    norm_eps: float = 1e-5
    rope_theta: float = 50000
    norm_type: str = "rmsnorm"
    rope_type: str = "default"
    hidden_act: str = "silu"
    hf_config: AutoConfig = None


class DeepseekV3RotaryEmbedding(nn.Module):
    def __init__(self, dim, max_position_embeddings=2048, base=10000, device=None):
        super().__init__()

        self.dim = dim
        self.max_position_embeddings = max_position_embeddings
        self.base = base
        self.max_seq_len_cached = None

    def reset_rotary_cache(self, device=None):
        if hasattr(self, "inv_freq"):
            return

        inv_freq = 1.0 / (
            self.base ** (torch.arange(0, self.dim, 2).float().to(device) / self.dim)
        )
        self.register_buffer("inv_freq", inv_freq, persistent=False)

        # Build here to make `torch.jit.trace` work.
        self._set_cos_sin_cache(
            seq_len=self.max_position_embeddings,
            device=self.inv_freq.device,
            dtype=torch.get_default_dtype(),
        )

    def _set_cos_sin_cache(self, seq_len, device, dtype):
        self.max_seq_len_cached = seq_len
        t = torch.arange(
            self.max_seq_len_cached, device=device, dtype=self.inv_freq.dtype
        )

        freqs = torch.outer(t, self.inv_freq.to(t.device))
        # Different from paper, but it uses a different permutation in order to obtain the same calculation
        emb = torch.cat((freqs, freqs), dim=-1)
        self.register_buffer("cos_cached", emb.cos().to(dtype), persistent=False)
        self.register_buffer("sin_cached", emb.sin().to(dtype), persistent=False)

    @torch.no_grad()
    def forward(self, x, seq_len=None):
        # x: [bs, num_attention_heads, seq_len, head_size]
        if self.max_seq_len_cached is None or seq_len > self.max_seq_len_cached:
            self._set_cos_sin_cache(seq_len=seq_len, device=x.device, dtype=x.dtype)

        return (
            self.cos_cached[:seq_len].to(dtype=x.dtype),
            self.sin_cached[:seq_len].to(dtype=x.dtype),
        )


def rotate_half(x):
    """Rotates half the hidden dims of the input."""
    x1 = x[..., : x.shape[-1] // 2]
    x2 = x[..., x.shape[-1] // 2 :]
    return torch.cat((-x2, x1), dim=-1)


def yarn_get_mscale(scale=1, mscale=1):
    if scale <= 1:
        return 1.0
    return 0.1 * mscale * math.log(scale) + 1.0


# Copied from transformers.models.llama.modeling_llama.apply_rotary_pos_emb
def apply_rotary_pos_emb(q, k, cos, sin, position_ids, unsqueeze_dim=1):
    """Applies Rotary Position Embedding to the query and key tensors.
    Args:
        q (`torch.Tensor`): The query tensor.
        k (`torch.Tensor`): The key tensor.
        cos (`torch.Tensor`): The cosine part of the rotary embedding.
        sin (`torch.Tensor`): The sine part of the rotary embedding.
        position_ids (`torch.Tensor`):
            The position indices of the tokens corresponding to the query and key tensors. For example, this can be
            used to pass offsetted position ids when working with a KV-cache.
        unsqueeze_dim (`int`, *optional*, defaults to 1):
            The 'unsqueeze_dim' argument specifies the dimension along which to unsqueeze cos[position_ids] and
            sin[position_ids] so that they can be properly broadcasted to the dimensions of q and k. For example, note
            that cos[position_ids] and sin[position_ids] have the shape [batch_size, seq_len, head_dim]. Then, if q and
            k have the shape [batch_size, heads, seq_len, head_dim], then setting unsqueeze_dim=1 makes
            cos[position_ids] and sin[position_ids] broadcastable to the shapes of q and k. Similarly, if q and k have
            the shape [batch_size, seq_len, heads, head_dim], then set unsqueeze_dim=2.
    Returns:
        `tuple(torch.Tensor)` comprising of the query and key tensors rotated using the Rotary Position Embedding.
    """
    cos = cos[position_ids].unsqueeze(unsqueeze_dim)
    sin = sin[position_ids].unsqueeze(unsqueeze_dim)

    b, h, s, d = q.shape
    q = q.view(b, h, s, d // 2, 2).transpose(4, 3).reshape(b, h, s, d)

    b, h, s, d = k.shape
    k = k.view(b, h, s, d // 2, 2).transpose(4, 3).reshape(b, h, s, d)

    q_embed = (q * cos) + (rotate_half(q) * sin)
    k_embed = (k * cos) + (rotate_half(k) * sin)
    return q_embed, k_embed


# Copied from transformers.models.llama.modeling_llama.repeat_kv
def repeat_kv(hidden_states: torch.Tensor, n_rep: int) -> torch.Tensor:
    """
    This is the equivalent of torch.repeat_interleave(x, dim=1, repeats=n_rep). The hidden states go from (batch,
    num_key_value_heads, seqlen, head_dim) to (batch, num_attention_heads, seqlen, head_dim)
    """
    batch, num_key_value_heads, slen, head_dim = hidden_states.shape
    if n_rep == 1:
        return hidden_states
    hidden_states = hidden_states[:, :, None, :, :].expand(
        batch, num_key_value_heads, n_rep, slen, head_dim
    )
    return hidden_states.reshape(batch, num_key_value_heads * n_rep, slen, head_dim)


# Copied from transformers.models.llama.modeling_llama.LlamaAttention with Llama->DeepseekV3
class DeepseekV3Attention(nn.Module):
    """Multi-headed attention from 'Attention Is All You Need' paper"""

    def __init__(self, config: DeepseekV3MoeArgs, layer_idx: Optional[int] = None):
        super().__init__()
        self.config = config
        self.layer_idx = layer_idx
        if layer_idx is None:
            logger.warning_once(
                f"Instantiating {self.__class__.__name__} without passing `layer_idx` is not recommended and will "
                "to errors during the forward call, if caching is used. Please make sure to provide a `layer_idx` "
                "when creating this class."
            )

        self.hidden_size = config.dim
        self.num_heads = config.n_heads
        self.max_position_embeddings = config.max_seq_len
        self.rope_theta = config.rope_theta
        self.q_lora_rank = config.q_lora_rank
        self.qk_rope_head_dim = config.qk_rope_head_dim
        self.kv_lora_rank = config.kv_lora_rank
        self.v_head_dim = config.v_head_dim
        self.qk_nope_head_dim = config.qk_nope_head_dim
        self.q_head_dim = config.qk_nope_head_dim + config.qk_rope_head_dim
        self.attn_func = None

        self.is_causal = True

        if self.q_lora_rank is None:
            self.q_proj = nn.Linear(
                self.hidden_size, self.num_heads * self.q_head_dim, bias=False
            )
        else:
            self.q_a_proj = nn.Linear(
                self.hidden_size, config.q_lora_rank, bias=config.attention_bias
            )
            self.q_a_layernorm = RMSNorm(
                config.q_lora_rank, casting_mode=config.hf_config.model_type
            )
            self.q_b_proj = nn.Linear(
                config.q_lora_rank, self.num_heads * self.q_head_dim, bias=False
            )

        self.kv_a_proj_with_mqa = nn.Linear(
            self.hidden_size,
            config.kv_lora_rank + config.qk_rope_head_dim,
            bias=config.attention_bias,
        )
        self.kv_a_layernorm = RMSNorm(
            config.kv_lora_rank, casting_mode=config.hf_config.model_type
        )
        self.kv_b_proj = nn.Linear(
            config.kv_lora_rank,
            self.num_heads
            * (self.q_head_dim - self.qk_rope_head_dim + self.v_head_dim),
            bias=False,
        )

        self.o_proj = nn.Linear(
            self.num_heads * self.v_head_dim,
            self.hidden_size,
            bias=config.attention_bias,
        )
        self._init_rope()

        self.softmax_scale = self.q_head_dim ** (-0.5)
        if self.config.rope_scaling is not None:
            mscale_all_dim = self.config.rope_scaling.get("mscale_all_dim", 0)
            scaling_factor = self.config.rope_scaling["factor"]
            if mscale_all_dim:
                mscale = yarn_get_mscale(scaling_factor, mscale_all_dim)
                self.softmax_scale = self.softmax_scale * mscale * mscale

    def _init_rope(self):
        if self.config.rope_scaling is None:
            self.rotary_emb = DeepseekV3RotaryEmbedding(
                self.qk_rope_head_dim,
                max_position_embeddings=self.max_position_embeddings,
                base=self.rope_theta,
            )
        else:
            scaling_type = self.config.rope_scaling["type"]
            # scaling_factor = self.config.rope_scaling["factor"]
            # if scaling_type == "linear":
            #     self.rotary_emb = DeepseekV3LinearScalingRotaryEmbedding(
            #         self.qk_rope_head_dim,
            #         max_position_embeddings=self.max_position_embeddings,
            #         scaling_factor=scaling_factor,
            #         base=self.rope_theta,
            #     )
            # elif scaling_type == "dynamic":
            #     self.rotary_emb = DeepseekV3DynamicNTKScalingRotaryEmbedding(
            #         self.qk_rope_head_dim,
            #         max_position_embeddings=self.max_position_embeddings,
            #         scaling_factor=scaling_factor,
            #         base=self.rope_theta,
            #     )
            # elif scaling_type == "yarn":
            #     kwargs = {
            #         key: self.config.rope_scaling[key]
            #         for key in [
            #             "original_max_position_embeddings",
            #             "beta_fast",
            #             "beta_slow",
            #             "mscale",
            #             "mscale_all_dim",
            #         ]
            #         if key in self.config.rope_scaling
            #     }
            #     self.rotary_emb = DeepseekV3YarnRotaryEmbedding(
            #         self.qk_rope_head_dim,
            #         max_position_embeddings=self.max_position_embeddings,
            #         scaling_factor=scaling_factor,
            #         base=self.rope_theta,
            #         **kwargs,
            #     )
            # else:
            raise ValueError(f"Unknown RoPE scaling type {scaling_type}")

    def _shape(self, tensor: torch.Tensor, seq_len: int, bsz: int):
        return (
            tensor.view(bsz, seq_len, self.num_heads, self.v_head_dim)
            .transpose(1, 2)
            .contiguous()
        )

    def forward(
        self,
        hidden_states: torch.Tensor,
        position_ids: torch.LongTensor,
    ):
        bsz, q_len, _ = hidden_states.size()

        if self.q_lora_rank is None:
            q = self.q_proj(hidden_states)
        else:
            q = self.q_b_proj(self.q_a_layernorm(self.q_a_proj(hidden_states)))

        q = q.view(bsz, q_len, -1, self.q_head_dim).transpose(1, 2)
        q_nope, q_pe = torch.split(
            q, [self.qk_nope_head_dim, self.qk_rope_head_dim], dim=-1
        )

        compressed_kv = self.kv_a_proj_with_mqa(hidden_states)
        compressed_kv, k_pe = torch.split(
            compressed_kv, [self.kv_lora_rank, self.qk_rope_head_dim], dim=-1
        )
        k_pe = k_pe.view(bsz, q_len, 1, self.qk_rope_head_dim).transpose(1, 2)
        kv = (
            self.kv_b_proj(self.kv_a_layernorm(compressed_kv))
            .view(bsz, q_len, -1, self.qk_nope_head_dim + self.v_head_dim)
            .transpose(1, 2)
        )

        k_nope, value_states = torch.split(
            kv, [self.qk_nope_head_dim, self.v_head_dim], dim=-1
        )
        kv_seq_len = value_states.shape[-2]
        self.rotary_emb.reset_rotary_cache(device=value_states.device)
        cos, sin = self.rotary_emb(value_states, seq_len=kv_seq_len)
        q_pe, k_pe = apply_rotary_pos_emb(q_pe, k_pe, cos, sin, position_ids)

        query_states = torch.cat([q_nope, q_pe], dim=-1)
        k_pe_expanded = k_pe.expand(-1, k_nope.shape[1], -1, -1)
        key_states = torch.cat([k_nope, k_pe_expanded], dim=-1)

        attn_output = F.scaled_dot_product_attention(
            query_states, key_states, value_states, is_causal=True
        )
        attn_output = attn_output.transpose(1, 2).contiguous()
        attn_output = attn_output.reshape(bsz, q_len, -1)
        return self.o_proj(attn_output)


class FakeLinear(nn.Module):
    def __init__(self, in_features: int, out_features: int, num_experts: int):
        super().__init__()
        self.weight = nn.Parameter(torch.empty(num_experts, out_features, in_features))


class DeepseekV3MLP(nn.Module):
    def __init__(self, config, hidden_size=None, intermediate_size=None):
        super().__init__()
        self.config = config
        self.hidden_size = config.dim if hidden_size is None else hidden_size
        self.intermediate_size = (
            config.intermediate_size if intermediate_size is None else intermediate_size
        )
        self.gate_proj = nn.Linear(self.hidden_size, self.intermediate_size, bias=False)
        self.up_proj = nn.Linear(self.hidden_size, self.intermediate_size, bias=False)
        self.down_proj = nn.Linear(self.intermediate_size, self.hidden_size, bias=False)
        self.act_mul_func = MLPActMulFunc(ACT2FN[config.hidden_act])

    def forward(self, x):
        return self.down_proj(self.act_mul_func(self.gate_proj(x), self.up_proj(x)))


class DeepseekV3MLPWithEP(nn.Module):
    def __init__(self, config, hidden_size=None, intermediate_size=None, experts_num=1):
        super().__init__()
        self.config = config
        self.experts_num = experts_num
        self.hidden_size = config.dim if hidden_size is None else hidden_size
        self.intermediate_size = (
            config.intermediate_size if intermediate_size is None else intermediate_size
        )
        # Transpose the MLP for group gemm
        self.gate_proj = FakeLinear(
            self.hidden_size, self.intermediate_size, self.experts_num
        )
        self.up_proj = FakeLinear(
            self.hidden_size, self.intermediate_size, self.experts_num
        )
        self.down_proj = FakeLinear(
            self.intermediate_size, self.hidden_size, self.experts_num
        )
        self.act_fn = ACT2FN[config.hidden_act]


class MoEGate(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.config = config
        self.top_k = config.hf_config.num_experts_per_tok
        self.n_routed_experts = config.n_routed_experts
        self.routed_scaling_factor = config.routed_scaling_factor
        self.scoring_func = config.scoring_func
        self.seq_aux = config.seq_aux
        self.topk_method = config.topk_method
        self.n_group = config.n_group
        self.topk_group = config.topk_group

        # topk selection algorithm
        self.norm_topk_prob = config.norm_topk_prob
        self.gating_dim = config.dim
        self.weight = nn.Parameter(
            torch.empty((self.n_routed_experts, self.gating_dim))
        )
        if self.topk_method == "noaux_tc":
            self.e_score_correction_bias = nn.Parameter(
                torch.empty((self.n_routed_experts))
            )

        self.reset_parameters()

    def reset_parameters(self) -> None:
        import torch.nn.init as init

        init.kaiming_uniform_(self.weight, a=math.sqrt(5))

    def forward(self, hidden_states):
        bsz, seq_len, h = hidden_states.shape
        ### compute gating score
        hidden_states = hidden_states.view(-1, h)
        logits = F.linear(
            hidden_states.type(torch.float32), self.weight.type(torch.float32), None
        )
        if self.scoring_func == "sigmoid":
            scores = logits.sigmoid()
        else:
            raise NotImplementedError(
                f"insupportable scoring function for MoE gating: {self.scoring_func}"
            )

        ### select top-k experts
        if self.topk_method == "noaux_tc":
            # assert not self.training
            scores_for_choice = scores.view(
                bsz * seq_len, -1
            ) + self.e_score_correction_bias.unsqueeze(0)
            group_scores = (
                scores_for_choice.view(bsz * seq_len, self.n_group, -1)
                .topk(2, dim=-1)[0]
                .sum(dim=-1)
            )  # [n, n_group]
            group_idx = torch.topk(
                group_scores, k=self.topk_group, dim=-1, sorted=False
            )[1]  # [n, top_k_group]
            group_mask = torch.zeros_like(group_scores)  # [n, n_group]
            group_mask.scatter_(1, group_idx, 1)  # [n, n_group]
            score_mask = (
                group_mask.unsqueeze(-1)
                .expand(
                    bsz * seq_len, self.n_group, self.n_routed_experts // self.n_group
                )
                .reshape(bsz * seq_len, -1)
            )  # [n, e]
            tmp_scores = scores_for_choice.masked_fill(
                ~score_mask.bool(), 0.0
            )  # [n, e]
            _, topk_idx = torch.topk(tmp_scores, k=self.top_k, dim=-1, sorted=False)
            topk_weight = scores.gather(1, topk_idx)
        else:
            raise NotImplementedError(
                f"insupportable TopK function for MoE gating: {self.topk_method}"
            )

        ### norm gate to sum 1
        if self.top_k > 1 and self.norm_topk_prob:
            denominator = topk_weight.sum(dim=-1, keepdim=True) + 1e-20
            topk_weight = topk_weight / denominator
        topk_weight = (
            topk_weight * self.routed_scaling_factor
        )  # must multiply the scaling factor

        return topk_idx, topk_weight


class DeepseekV3MoE(nn.Module):
    """
    A mixed expert module containing shared experts.
    """

    token_send_buf: Optional[torch.Tensor] = None
    token_gather_buf: Optional[torch.Tensor] = None

    def __init__(self, config):
        super().__init__()
        self.config = config
        self.num_experts_per_tok = config.hf_config.num_experts_per_tok
        self.local_experts = config.n_routed_experts
        self.total_experts = config.n_routed_experts
        self.local_to_dtensor = IdentityLayer()
        self.reshard_helper_layer = IdentityLayer()
        self.gate = MoEGate(config)
        self.ep_size = 1
        self.experts_per_rank = config.n_routed_experts
        self.experts = nn.ModuleList(
            [
                DeepseekV3MLP(config, intermediate_size=config.moe_intermediate_size)
                for i in range(config.n_routed_experts)
            ]
        )
        self.experts_ep = DeepseekV3MLPWithEP(
            config,
            intermediate_size=config.moe_intermediate_size,
            experts_num=config.n_routed_experts,
        )
        if config.n_shared_experts is not None:
            intermediate_size = config.moe_intermediate_size * config.n_shared_experts
            self.shared_experts = DeepseekV3MLP(
                config=config, intermediate_size=intermediate_size
            )
        self.group_gemm_imp = group_gemm_imp()

    def sort_tokens(self, x, topk_ids, topk_weights):
        # This part sorts the token indices so that tokens routed to the same expert reside consecutively.
        # An implication is that tokens to the same "expert group" (i.e., device) are also consecutive.
        # Since this is an "aritificial" index creation (final outcome being
        # `idxs`), we don't need gradients here.

        with torch.no_grad():
            # [seq_len, n_routed_experts]
            expert_counts = topk_ids.new_zeros((topk_ids.shape[0], self.total_experts))
            # Fill 1 to the selected experts
            expert_counts.scatter_(1, topk_ids, 1)
            tokens_per_expert = expert_counts.sum(dim=0)
            # Token indices for each expert
            token_indices = topk_ids.view(-1).argsort()

        sorted_tokens = x[token_indices // topk_ids.shape[1]]
        # assert sorted_tokens.shape == sorted_tokens_shape

        return (sorted_tokens, token_indices, tokens_per_expert)

    def get_send_buf(self):
        # [Why detach?] During a first forward-backward step, the buffer would
        # be included in a computational graph. In a second step, autograd will
        # return an error saying "Trying to backward through the graph a second
        # time (or directly access saved tensors more than once)". This is
        # because the buffer is still in the graph, and autograd is trying to
        # backward through the graph a second time. To avoid this, we detach the
        # buffer from the graph. `detach()` returns a new tensor, which shares
        # the same storage with the original one.
        self.token_send_buf.grad = None
        return self.token_send_buf.detach()

    def get_gather_buf(self):
        # See [Why detach?] in `get_send_buf`
        self.token_gather_buf.grad = None
        return self.token_gather_buf.detach()

    def moe_infer_ep(self, x, topk_ids, topk_weight):
        """
        x: [batch * local_seq_len, dim]
        topk_ids: [batch * local_seq_len, topk]
        topk_weight: [batch * local_seq_len, topk]

        sorted_tokens: [batch * local_seq_len * topk, dim]
        token_indices: [batch * local_seq_len * topk]
        tokens_per_expert: [n_experts]
        """
        (
            sorted_tokens,
            token_indices,
            tokens_per_expert,
        ) = self.sort_tokens(x, topk_ids, topk_weight)
        # keep the seqlen dimension for later use without holding onto the sorted tokens
        seqlen_sorted_tokens = sorted_tokens.shape[0]

        # Sum the tokens over local experts, then we get tokens per EP rank,
        # which is the input splits
        with torch.no_grad():
            # tokens_per_expert: [n_experts, 1]
            # tokens_per_expert_group: [n_experts, 1]
            tokens_per_expert_group = tokens_per_expert.new_empty(
                tokens_per_expert.shape[0]
            )

            # For TP/EP mode, the input is sequencely parallelized
            # So each EP group will have distinct, but the same number of tokens
            # After this collective, tokens_per_expert_group is still of shape [n_experts, 1]

            # Let's say we are on EP rank 0:
            # recv: [(e0, e1, e2 ...), (e0, e1, e2 ...), ...], totally `n_experts` elements
            #        ----------------: tokens from EP group 0 to EP group 0
            #                          ----------------: tokens from EP group 1 to EP group 0
            #                          ...
            # So we can just concat
            dist.all_to_all_single(
                tokens_per_expert_group,
                tokens_per_expert,
                group=self.ep_group,
                async_op=False,
            )
            input_splits = tokens_per_expert.view(self.ep_size, -1).sum(dim=1)
        # Move input to the `token_send_buf` symm mem
        token_send_buf = self.get_send_buf()
        token_send_buf[: token_indices.shape[0]].copy_(sorted_tokens)
        # Note: `out=` avoids copy, but it is not differentiable
        # torch.index_select(x, 0, idxs // topk_ids.shape[1], out=token_send_buf[: idxs.shape[0]])

        # Reference:
        #   1. [TorchTitan](https://github.com/pytorch/torchtitan/blob/main/torchtitan/experiments/deepseek_v3/symm_mem_recipes/triton_on_device_all_to_all_v.py)
        #   2. [Symm-mem-recipes](https://github.com/yifuwang/symm-mem-recipes)
        token_gather_buf, output_splits = OnDeviceAllToAllV.apply(
            token_send_buf,
            input_splits,
            self.ep_group,
        )

        # We need to permute the received tokens so that tokens for the same expert are contiguous.
        # This part prepares a 1D tensor `permuted_indices` for such permutation.
        # This part doesn't need gradient.
        with torch.no_grad():
            ALIGN_SIZE_M = 128
            permuted_indices, m_sizes, m_offsets = generate_permute_indices(
                tokens_per_expert_group,
                self.local_experts,
                self.ep_size,
                token_gather_buf.shape[0],
                ALIGN_SIZE_M,
            )
        # Permute the received tokens so that tokens for the same expert are contiguous.
        contig_tokens = token_gather_buf[permuted_indices]

        # group gemm - handle all three group gemms (up, gate, down for all experts)
        # print(f"m_sizes: {m_sizes}, m_offsets: {m_offsets}")
        hidden_outputs = self.group_gemm_imp(
            contig_tokens,
            m_sizes,
            m_offsets,
            self.experts_ep.gate_proj.weight.to_local(),
            self.experts_ep.up_proj.weight.to_local(),
            self.experts_ep.down_proj.weight.to_local(),
            self.experts_ep.act_fn,
        )

        # Prepare buffer for tokens processed by experts
        processed_tokens = self.get_gather_buf()

        # Move into Symmetric Memory for the return shuffle
        processed_tokens[permuted_indices] = hidden_outputs

        # Now shuffle the tokens back to their original owner, i.e. EP to DP shuffle.
        # The input/output splits are just a reverse of the previous shuffle.
        token_return_buf, _ = OnDeviceAllToAllV.apply(
            processed_tokens,
            output_splits,
            self.ep_group,
        )

        returned_tokens = token_return_buf[:seqlen_sorted_tokens]
        output_tokens = torch.empty_like(returned_tokens)
        output_tokens[token_indices] = returned_tokens

        final_out = (
            output_tokens.view(*topk_ids.shape, -1)
            .type(topk_weight.dtype)
            .mul_(topk_weight.unsqueeze(dim=-1))
            .sum(dim=1)
            .type(returned_tokens.dtype)
        )

        return final_out

    def moe_infer(self, x, topk_ids, topk_weight):
        cnts = topk_ids.new_zeros((topk_ids.shape[0], self.config.n_routed_experts))
        cnts.scatter_(1, topk_ids, 1)
        tokens_per_expert = cnts.sum(dim=0)
        idxs = topk_ids.view(-1).argsort()
        sorted_tokens = x[idxs // topk_ids.shape[1]]
        tokens_per_expert = tokens_per_expert.cpu().numpy()
        outputs = []
        start_idx = 0
        for i, num_tokens in enumerate(tokens_per_expert):
            end_idx = start_idx + num_tokens
            if num_tokens == 0:
                continue
            expert = self.experts[i]
            tokens_for_this_expert = sorted_tokens[start_idx:end_idx]
            expert_out = expert(tokens_for_this_expert)
            outputs.append(expert_out)
            start_idx = end_idx

        outs = torch.cat(outputs, dim=0) if len(outputs) else sorted_tokens.new_empty(0)

        new_x = torch.empty_like(outs)
        new_x[idxs] = outs
        final_out = (
            new_x.view(*topk_ids.shape, -1)
            .type(topk_weight.dtype)
            .mul_(topk_weight.unsqueeze(dim=-1))
            .sum(dim=1)
            .type(new_x.dtype)
        )
        return final_out

    def forward(self, hidden_states):
        identity = hidden_states
        orig_shape = hidden_states.shape
        topk_idx, topk_weight = self.gate(hidden_states)
        hidden_states = hidden_states.view(-1, hidden_states.shape[-1])
        if self.ep_size > 1:
            y = self.moe_infer_ep(hidden_states, topk_idx, topk_weight).view(
                *orig_shape
            )
        else:
            y = self.moe_infer(hidden_states, topk_idx, topk_weight).view(*orig_shape)
        # y = self.moe_infer(hidden_states, topk_idx, topk_weight).view(*orig_shape)
        if self.config.n_shared_experts is not None:
            y = y + self.shared_experts(identity)
        return self.reshard_helper_layer(y)


class DeepseekV3DecoderLayer(nn.Module):
    def __init__(self, config: DeepseekV3MoeArgs, layer_idx: int):
        super().__init__()
        self.hidden_size = config.dim
        self.layer_idx = layer_idx
        self.self_attn = DeepseekV3Attention(config=config, layer_idx=layer_idx)
        self.use_moe = (
            config.n_routed_experts is not None
            and layer_idx >= config.first_k_dense_replace
            and layer_idx % config.moe_layer_freq == 0
        )

        self.mlp = DeepseekV3MoE(config) if self.use_moe else DeepseekV3MLP(config)

        self.input_layernorm = RMSNorm(
            config.dim, eps=config.norm_eps, casting_mode=config.hf_config.model_type
        )
        self.post_attention_layernorm = RMSNorm(
            config.dim, eps=config.norm_eps, casting_mode=config.hf_config.model_type
        )

    def forward(
        self,
        hidden_states: torch.Tensor,
        position_ids: Optional[torch.LongTensor] = None,
    ):
        """
        Args:
            hidden_states (`torch.FloatTensor`): input to the layer of shape `(batch, seq_len, embed_dim)`
        """
        # Self Attention
        h = hidden_states + self.self_attn(
            self.input_layernorm(hidden_states),
            position_ids,
        )
        # Fully Connected
        output = self.mlp(self.post_attention_layernorm(h))
        output = h + output
        return output


class DeepseekV3MoEModel(BaseModel):
    """
    DeepseekV3MoEModel Module

    Args:
        model_args (DeepseekV3MoeArgs): Model configuration arguments.

    Attributes:
        model_args (DeepseekV3MoeArgs): Model configuration arguments.
        vocab_size (int): Vocabulary size.
        n_layers (int): Number of layers in the model.
        embed_tokens (ParallelEmbedding): Token embeddings.
        layers (torch.nn.ModuleList): List of DeepseekV3MoE blocks.
        norm (RMSNorm): Layer normalization for the model output.
        lm_head (ColumnParallelLinear): Linear layer for final output.
    """

    @staticmethod
    def supported_model_types():
        return ["deepseek_v3"]

    def __init__(self, model_args: DeepseekV3MoeArgs):
        super().__init__(model_args.hf_config)
        self.model_args = model_args
        self.vocab_size = model_args.vocab_size
        self.n_layers = model_args.n_layers
        self.embed_tokens = nn.Embedding(model_args.vocab_size, model_args.dim)

        self.layers = nn.ModuleList(
            [
                DeepseekV3DecoderLayer(model_args, layer_idx)
                for layer_idx in range(model_args.n_layers)
            ]
        )
        self.norm = RMSNorm(
            model_args.dim,
            model_args.norm_eps,
            casting_mode=model_args.hf_config.model_type,
        )

        if not model_args.hf_config.tie_word_embeddings:
            self.tie_embed_tokens = False
            self.lm_head = nn.Linear(
                model_args.dim,
                model_args.vocab_size,
                bias="lm_head" in model_args.biases,
            )
        else:
            self.tie_embed_tokens = True
        self.identity_layer = IdentityLayer()

    def forward(
        self,
        input_ids: torch.Tensor,
        position_ids: torch.Tensor = None,
        *args,
        **kwargs,
    ):
        if self.embed_tokens is not None:
            hidden_states = self.embed_tokens(input_ids)
            # Do not remove this line
            # This is a trick for TP with torch.compile
            hidden_states = self.identity_layer(hidden_states)
        else:
            hidden_states = input_ids

        for decoder_layer in self.layers:
            if (
                hasattr(decoder_layer, "_gradient_checkpointing_enabled")
                and decoder_layer._gradient_checkpointing_enabled
            ):
                hidden_states = torch.utils.checkpoint.checkpoint(
                    decoder_layer,
                    hidden_states,
                    position_ids,
                    use_reentrant=False,
                )
            else:
                hidden_states = decoder_layer(
                    hidden_states,
                    position_ids=position_ids,
                )

        # Add `if` check just in case `pp` is enabled
        if self.norm is not None:
            hidden_states = self.norm(hidden_states)
            if not self.tie_embed_tokens:
                output = self.lm_head(hidden_states)
            else:
                is_w_dist_tensor = isinstance(
                    self.embed_tokens.weight, torch.distributed.tensor.DTensor
                )
                embed_tokens_weight = (
                    self.embed_tokens.weight.full_tensor()
                    if is_w_dist_tensor
                    else self.embed_tokens.weight
                )
                is_a_dist_tensor = isinstance(
                    hidden_states, torch.distributed.tensor.DTensor
                )
                hidden_states = (
                    hidden_states.full_tensor() if is_a_dist_tensor else hidden_states
                )
                output = hidden_states @ embed_tokens_weight.t()
            return output
        else:
            return hidden_states

    def post_to_empty_hook(self, cosmos_config: CosmosConfig):
        for layer in self.layers:
            if layer.use_moe:
                layer.mlp.gate.weight.requires_grad_(False)

        # rotary.inv_freq could get deleted and not re-initialized
        # so we need to delete it manually
        # self.rotary_emb.reset_inv_freq()
        # Basically, max_seq_len * 2 is enough for all-to-all-v communication.
        overflow = 2

        # TODO(cjx): max_seq_len * mini_batch is a better choice
        MAX_BATCH_MUL_SEQ_LEN = (
            self.model_args.max_seq_len
            * cosmos_config.train.train_batch_per_replica
            * self.model_args.hf_config.num_experts_per_tok
        )

        OnDeviceAllToAllV.max_output_len = MAX_BATCH_MUL_SEQ_LEN * overflow
        # Init MoE kernel related buffers
        if DeepseekV3MoE.token_send_buf is None:
            dtype = self.model_args.hf_config.torch_dtype

            # Input buffer for DP-to-EP shuffle
            DeepseekV3MoE.token_send_buf = symm_mem.empty(
                MAX_BATCH_MUL_SEQ_LEN,
                self.model_args.dim,  # hidden dim
                dtype=dtype,
                device=self.current_device(),
            )
            DeepseekV3MoE.token_send_buf.zero_()
            # Input buffer for EP-to-DP shuffle
            DeepseekV3MoE.token_gather_buf = symm_mem.empty(
                MAX_BATCH_MUL_SEQ_LEN * overflow,
                self.model_args.dim,  # hidden dim
                dtype=dtype,
                device=self.current_device(),
            )
            DeepseekV3MoE.token_gather_buf.zero_()

    @property
    def parallelize_fn(self):
        from .parallelize import parallelize

        return parallelize, self

    def apply_pipeline_split(self, pp_rank, pp_size):
        """
        Apply pipeline split to the model.
        This typically involves splitting the model into multiple stages,
        and moving each stage to a different device.
        """
        assert pp_size > 1
        is_first = pp_rank == 0
        is_last = pp_rank == pp_size - 1

        # Compute the layers belonging to this stage
        n_layers = len(self.layers)
        layers_per_stage = n_layers // pp_size

        if not is_first:
            self.embed_tokens = None
        if not is_last:
            self.lm_head = None
            self.norm = None

        local_layers = torch.nn.ModuleDict()
        for i in range(
            pp_rank * layers_per_stage,
            ((pp_rank + 1) * layers_per_stage) if not is_last else n_layers,
        ):
            local_layers[str(i)] = self.layers[str(i)]

        # Reset the layers for pipeline splitting
        self.layers = local_layers

    def load_hf_weights(
        self,
        model_name_or_path: str,
        parallel_dims: ParallelDims,
        device: torch.device,
        revision: Optional[str] = None,
    ):
        """
        Load weights from a HuggingFace model.

        Args:
            model_path (str): Path to the HuggingFace model.
            parallel_dims (ParallelDims): Parallel dimensions definition.
            device (torch.device): Device to load the weights.
        """
        # Load all safetensors from `model_path`
        model_type = retry(AutoConfig.from_pretrained)(model_name_or_path).model_type
        model_path = resolve_model_path(model_name_or_path, revision=revision)
        safetensors_files = [
            f for f in os.listdir(model_path) if f.endswith(".safetensors")
        ]

        self_state_dict = self.state_dict()
        self_state_dict = {clear_weight_name(k): v for k, v in self_state_dict.items()}

        lm_head_weight_key = "lm_head.weight"
        embed_tokens_weight_key = "model.embed_tokens.weight"
        weights_of_ckpt_names = set()
        reserved = {}
        _, tp_ep_size = parallel_dims.tp_coord

        for f in safetensors_files:
            weights_of_ckpt = {}
            ckpt = safe_open(
                os.path.join(model_path, f), framework="pt", device=str(device)
            )
            keys = ckpt.keys()
            for name in keys:
                ckpt_tensor = ckpt.get_tensor(name)
                weights_of_ckpt[name] = ckpt_tensor
                weights_of_ckpt_names.add(name)
                if name == embed_tokens_weight_key:
                    reserved[name] = ckpt_tensor

            for name in weights_of_ckpt.keys():
                tensor = weights_of_ckpt[name]
                dest_name, shared_weight = convert_weight_from_hf(
                    tensor,
                    name,
                    model_type,
                    parallel_dims,
                    n_experts=self.model_args.n_routed_experts,
                )

                if dest_name is None:
                    # This is due to the expert parallelism grouping
                    continue

                expert_id = None
                if match := re.search(  # noqa: F841
                    r"layers\.(\d+)\.mlp\.experts\.(\d+)\.(up_proj|gate_proj|down_proj)\.(weight|bias)",
                    dest_name,
                ):
                    expert_id = int(match.group(2))
                    if tp_ep_size > 1:
                        dest_name = dest_name.replace(
                            f"experts.{expert_id}.", "experts_ep."
                        )

                    # Convert expert_id to local_expert_id
                    n_local_experts = (
                        self.model_args.n_routed_experts
                        // parallel_dims.tp
                        // parallel_dims.dp_shard
                    )

                    expert_id = expert_id % n_local_experts

                if dest_name not in self_state_dict and parallel_dims.pp_enabled:
                    logger.info(
                        f"Weight `{dest_name}` is discarded, maybe due to pipeline parallelism or expert parallelism grouping. Skipping this weight checking"
                    )
                    continue

                target_tensor = self_state_dict[dest_name]
                if isinstance(target_tensor, torch.distributed.tensor.DTensor):
                    target_tensor = target_tensor.to_local()
                # Write to the correct expert of the target tensor
                if expert_id is not None and tp_ep_size > 1:
                    target_tensor = target_tensor[expert_id]

                assert (
                    target_tensor.shape == shared_weight.shape
                ), f"Shape mismatch: {target_tensor.shape} != {shared_weight.shape} for {dest_name}"
                with torch.no_grad():
                    target_tensor.data.copy_(shared_weight)

        if (
            lm_head_weight_key not in weights_of_ckpt_names
            and embed_tokens_weight_key in weights_of_ckpt_names
        ):
            # tied with embed_tokens.weight
            name = lm_head_weight_key
            assert embed_tokens_weight_key in reserved
            tensor = reserved[embed_tokens_weight_key]
            dest_name, shared_weight = convert_weight_from_hf(
                tensor, name, model_type, parallel_dims
            )
            if dest_name in self_state_dict:
                target_tensor = self_state_dict[dest_name]
                is_dist_tensor = isinstance(
                    target_tensor, torch.distributed.tensor.DTensor
                )
                local_view = (
                    target_tensor.to_local() if is_dist_tensor else target_tensor
                )
                assert (
                    local_view.shape == shared_weight.shape
                ), f"Shape mismatch: {local_view.shape} != {shared_weight.shape} for {dest_name}"
                with torch.no_grad():
                    local_view.data.copy_(shared_weight)

    def get_position_ids(self, **kwargs) -> Tuple[torch.Tensor, int]:
        seq_dim_idx = 1
        inputs = kwargs["input_ids"]
        position_ids = (
            torch.arange(inputs.size(-1), dtype=torch.long, device=inputs.device)
            .unsqueeze(0)
            .expand_as(inputs)
        )
        return position_ids, inputs, seq_dim_idx

    def separate_model_parts(self) -> List[nn.Module]:
        return [self]

    @cached_property
    def _get_nparams_and_flops_fn(self) -> Callable[[int], tuple[int, int]]:
        nparams = sum(p.numel() for p in self.parameters())
        nparams_embedding = sum(
            sum(p.numel() for p in m.parameters())
            for m in self.children()
            if isinstance(m, nn.Embedding)
        )

        # Reasoning behind the factor of 12 for the self-attention part of the formula:
        # 1. each self-attention has 2 matmul in the forward and 4 in the backward (6)
        # 2. the flash attention does 1 more matmul recomputation in the backward
        #    but recomputation should not be counted in calculating MFU           (+0)
        # 3. each matmul performs 1 multiplication and 1 addition                 (*2)
        # 4. we follow the convention and do not account for sparsity in causal attention
        layers, heads, head_dim = (
            len(self.layers),
            self.model_args.n_heads,
            self.model_args.dim // self.model_args.n_heads,
        )
        return lambda seq_len: (
            nparams,
            6 * (nparams - nparams_embedding)
            + 12 * layers * heads * head_dim * seq_len,
        )

    def get_nparams_and_flops(self, seq_len: int) -> tuple[int, int]:
        return self._get_nparams_and_flops_fn(seq_len)

    @classmethod
    def from_model_args(cls, model_args: DeepseekV3MoeArgs) -> "DeepseekV3MoEModel":
        """
        Initialize a DeepseekV3MoE model from a DeepseekV3MoeArgs object.

        Args:
            model_args (DeepseekV3MoeArgs): Model configuration arguments.

        Returns:
            DeepseekV3MoE: DeepseekV3MoE model.

        """
        return cls(model_args)

    @classmethod
    def from_pretrained(
        cls,
        hf_config: AutoConfig,
        model_name_or_path: str,
        max_position_embeddings: Optional[int] = None,
    ) -> "DeepseekV3MoEModel":
        """
        Initialize a DeepseekV3MoE model from a pretrained model.

        Args:
            model_name_or_path (str): Model name or path to the pretrained model.

        Returns:
            DeepseekV3MoE: DeepseekV3MoE model.

        """
        try:
            if hf_config.model_type not in cls.supported_model_types():
                raise ValueError(f"Unsupported model type: {hf_config.model_type}")

            if max_position_embeddings is None:
                max_position_embeddings = hf_config.max_position_embeddings
            else:
                hf_config.max_position_embeddings = max_position_embeddings

            vocab_size = sync_model_vocab(model_name_or_path)
            rope_scaling = None
            if (
                hasattr(hf_config, "rope_scaling")
                and hf_config.rope_scaling is not None
            ):
                rope_scaling = hf_config.rope_scaling or {}
            rope_type = "default"
            if rope_scaling:
                rope_type = rope_scaling.get("rope_type", "default")

            # DeepseekV3MoE does not have any biases
            bias_list = []
            try:
                head_dim = hf_config.head_dim
            except Exception:
                head_dim = hf_config.hidden_size // hf_config.num_attention_heads
                logger.warning(f"head_dim not found in config, using {head_dim}")

            model = cls.from_model_args(
                DeepseekV3MoeArgs(
                    dim=hf_config.hidden_size,
                    intermediate_size=hf_config.intermediate_size,
                    moe_intermediate_size=hf_config.moe_intermediate_size,
                    n_layers=hf_config.num_hidden_layers,
                    n_heads=hf_config.num_attention_heads,
                    n_kv_heads=hf_config.num_key_value_heads,
                    head_dim=head_dim,
                    vocab_size=vocab_size,
                    max_seq_len=max_position_embeddings,
                    q_lora_rank=hf_config.q_lora_rank,
                    kv_lora_rank=hf_config.kv_lora_rank,
                    v_head_dim=hf_config.v_head_dim,
                    qk_nope_head_dim=hf_config.qk_nope_head_dim,
                    qk_rope_head_dim=hf_config.qk_rope_head_dim,
                    n_routed_experts=hf_config.n_routed_experts,
                    n_shared_experts=hf_config.n_shared_experts,
                    routed_scaling_factor=hf_config.routed_scaling_factor,
                    scoring_func=hf_config.scoring_func,
                    seq_aux=hf_config.seq_aux,
                    topk_method=hf_config.topk_method,
                    n_group=hf_config.n_group,
                    topk_group=hf_config.topk_group,
                    first_k_dense_replace=hf_config.first_k_dense_replace,
                    moe_layer_freq=hf_config.moe_layer_freq,
                    rope_theta=hf_config.rope_theta,
                    rope_type=rope_type,
                    rope_scaling=rope_scaling,
                    biases=bias_list,
                    attention_bias=hf_config.attention_bias,
                    norm_topk_prob=hf_config.norm_topk_prob,
                    hf_config=hf_config,
                )
            )

        except Exception as e:
            import traceback

            traceback.print_exc()
            raise e
        return model

    def check_cp_compatible(self, cp_size: int, tp_size: int):
        raise NotImplementedError(
            "Context Parallel is not supported for DeepseekV3MoEModel now."
        )
        # if not (self.model_args.n_heads % (cp_size * tp_size) == 0):
        #     raise ValueError(
        #         f"Model is not compatible with cp parallelism, model's head number={self.model_args.n_heads} is not divisible by cp size({cp_size}) * tp_size({tp_size}) = {cp_size * tp_size}"
        #     )



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cosmos_rl.utils.logging import logger
from cosmos_rl.utils.parallelism import ParallelDims
from cosmos_rl.utils.distributed import (
    init_distributed,
    destroy_distributed,
    get_controller_metadata,
)
from cosmos_rl.policy.trainer.sft_trainer import SFTTrainer
from cosmos_rl.policy.trainer.grpo_trainer import GRPOTrainer
from cosmos_rl.policy.config import Config as CosmosConfig
import torch
from cosmos_rl.utils import util


def main(*args, **kwargs):
    torch.backends.cuda.matmul.allow_bf16_reduced_precision_reduction = False
    ctrl_ip, ctrl_port, metadata = get_controller_metadata()

    if metadata["config"] is None:
        raise RuntimeError(
            f"[Policy] Please first go to http://{ctrl_ip}:{ctrl_port} to configure training parameters."
        )

    cosmos_config = CosmosConfig.from_dict(metadata["config"])

    logger.info(f"[Policy] Loaded configuration: {cosmos_config.model_dump()}")

    parallel_dims = ParallelDims.from_config(
        parallesim_config=cosmos_config.policy.parallelism
    )
    init_distributed()
    parallel_dims.build_mesh(device_type="cuda")

    policy_type = cosmos_config.train.train_policy.type

    try:
        with torch.autocast(
            device_type="cuda",
            dtype=util.str2torch_dtype(cosmos_config.train.param_dtype),
        ):
            if policy_type == "grpo":
                logger.info("Starting GRPO training...")
                trainer = GRPOTrainer(config=cosmos_config, parallel_dims=parallel_dims)
                trainer.main_loop()
            elif policy_type == "sft":
                logger.info("Starting SFT training...")
                trainer = SFTTrainer(config=cosmos_config, parallel_dims=parallel_dims)
                trainer.train()
            else:
                raise ValueError(f"Unknown policy type: {policy_type}")
    except Exception as e:
        import traceback

        traceback.print_exc()
        raise e
    finally:
        destroy_distributed()
        logger.info("Process group destroyed.")


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import argparse
import toml

from torch.utils.data import Dataset
from datasets import concatenate_datasets
import cosmos_rl.utils.util as util
import cosmos_rl.utils.cache as cache
from transformers import AutoTokenizer
from cosmos_rl.dispatcher.run_web_panel import main as launch_dispatcher
from cosmos_rl.policy.config import (
    Config,
    config_hash,
)


# This dataset is used for SFT with raw text input, which is used for models like Mistral
# It converts the conversation list to string format for models requiring raw text input
# This handles cases like Mistral where conversation dicts need string conversion
# to avoid role alternation errors
class SFTRawTextDataset(Dataset):
    def __init__(self, dataset: Dataset):
        self.dataset = dataset

    def setup(
        self,
        config: Config,
        tokenizer: AutoTokenizer,
    ):
        self.config = config.train.train_policy
        self.tokenizer = tokenizer
        self.column_name = self.config.conversation_column_name
        self.cache = None
        if self.config.enable_dataset_cache:
            cache_folder = os.path.join(
                os.environ.get(
                    "COSMOS_CACHE",
                    os.path.join(os.path.expanduser("~"), ".cache/cosmos/"),
                ),
                "datasets_cache",
                f"{self.config.dataset.name}-{config_hash(config)}",
            )
            print(f"SFTRawTextDataset Cache folder: {cache_folder}")
            self.cache = cache.DiskCache(cache_folder)

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx):
        # Check cache first if enabled
        if self.cache is not None:
            cached_item = self.cache.get(idx)
            if cached_item is not None:
                return cached_item

        # Retrieve raw item from dataset
        raw_item = (
            self.dataset[idx][self.column_name]
            if self.column_name
            else self.dataset[idx]
        )

        # Convert conversation list to string format
        if isinstance(raw_item, list):
            raw_item = "\n".join(
                [f"{turn['role']}: {turn['content']}" for turn in raw_item]
            )

        # Cache the processed item if caching is enabled
        if self.cache is not None:
            self.cache.set(idx, raw_item)

        return raw_item


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--config", type=str, required=True)
    args = parser.parse_known_args()[0]
    with open(args.config, "r") as f:
        config = toml.load(f)
    config = Config.from_dict(config)
    # Download HF dataset only on launcher worker
    dataset = util.load_data_from_disk_or_hf(
        config.train.train_policy.dataset.name,
        config.train.train_policy.dataset.subset,
        config.train.train_policy.dataset.revision or None,
    )
    dataset_list = []
    for split_name in config.train.train_policy.dataset.split:
        print(
            f"Appending split {split_name}, dataset size = {len(dataset[split_name])}"
        )
        dataset_list.append(dataset[split_name])
    train_dataset = concatenate_datasets(dataset_list)
    launch_dispatcher(
        dataset=SFTRawTextDataset(dataset=train_dataset),
    )



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Tuple, Optional

from cosmos_rl.utils.logging import logger

import tensorrt_llm

trtllm_version = tensorrt_llm.__version__

if trtllm_version == "1.0.0rc6":
    from tensorrt_llm import LLM
else:
    raise NotImplementedError(f"Unsupported trtllm version: {trtllm_version}")

from tensorrt_llm import SamplingParams
from tensorrt_llm.llmapi.llm_args import KvCacheConfig

from cosmos_rl.rollout.rollout_base import RolloutBase
from cosmos_rl.policy.config import Config
import cosmos_rl.utils.util as util
from cosmos_rl.dispatcher.data.packer import DataPacker

from transformers import AutoTokenizer
from transformers import AutoConfig
from transformers import GenerationConfig


class TRTLLM_Rollout(RolloutBase):
    def __init__(self, config: Config, tokenizer: AutoTokenizer, **kwargs):
        super().__init__(config, tokenizer)
        self.rollout_config = self.config.rollout

        hf_config_path = self.config.policy.model_name_or_path
        try:
            generation_config = util.retry(GenerationConfig.from_pretrained)(
                hf_config_path
            )
            self.eos_token_ids = generation_config.eos_token_id
            if isinstance(self.eos_token_ids, int):
                self.eos_token_ids = [self.eos_token_ids]
        except Exception as e:
            logger.warning(
                f"[Rollout] Failed to load generation config from {hf_config_path}: {str(e)}, use default eos_token_id."
            )
            # self.eos_token_ids = [tokenizer.eos_token_id]
            # TODO(lms): remove this
            self.eos_token_ids = [151645, 151643]

        self.model_config = util.retry(AutoConfig.from_pretrained)(
            self.config.policy.model_name_or_path
        )
        self._engine_initialized = True

    def init_engine(
        self,
        quantization: Optional[str] = None,
        seed: int = 42,
        load_format: str = "dummy",
        **kwargs,
    ):
        model_path = self.config.policy.model_name_or_path
        rollout_parallelism = self.rollout_config.parallelism

        tp_size = rollout_parallelism.tp_size
        pp_size = rollout_parallelism.pp_size
        assert pp_size == 1, "TRTLLM only support pipeline parallelism 1 now."
        assert quantization is None, "TRTLLM Rollout does not support quantization now."

        moe_tensor_parallel_size = -1  # by default, trtllm uses tp for MoE.

        # Check if the model has MoE
        moe_model_type = {"qwen3_moe"}
        # multimodal_type = {"qwen2_5_vl"}

        model_type = self.model_config.model_type
        if model_type in moe_model_type:
            moe_tensor_parallel_size = tp_size

        kwargs = {}
        kwargs["disable_overlap_scheduler"] = True
        policy_config = self.config.policy
        # Check the prefix_caching like arguments default enabled?
        kv_cache_config = KvCacheConfig(
            free_gpu_memory_fraction=self.rollout_config.gpu_memory_utilization,
            enable_block_reuse=False,  # disable KV cacheblock reuse for trtllm
        )
        self.rollout_engine = LLM(
            model=model_path,
            tensor_parallel_size=tp_size,
            pipeline_parallel_size=pp_size,
            dtype="auto",
            backend="pytorch",
            skip_tokenizer_init=False,
            max_seq_len=policy_config.model_max_length,
            max_num_tokens=2048
            if 2048 >= policy_config.model_max_length
            else policy_config.model_max_length,
            enable_chunked_prefill=self.rollout_config.enable_chunked_prefill,
            load_format=load_format,
            trust_remote_code=True,
            moe_expert_parallel_size=moe_tensor_parallel_size,
            kv_cache_config=kv_cache_config,
            # For cosmos-rl
            cosmos_config=self.config,
            **kwargs,
        )
        logger.info("[Rollout] TRTLLM Engine initialized.")

    def rollout_generation(
        self,
        prompt_id_and_payload_list: List[Tuple[int, str]],
        data_packer: DataPacker,
        sampling_params: SamplingParams,
    ) -> List[List[str]]:
        if not self._engine_initialized:
            raise RuntimeError(
                "[Rollout] Engine is not initialized, please call init_engine first."
            )

        # List of payloads.
        # [
        #   payload,
        #   payload,
        #   ...
        # ]
        payloads = [x[1] for x in prompt_id_and_payload_list]

        # Pack the payloads into prompts for vllm.
        prompts = [data_packer.get_rollout_input(payload) for payload in payloads]
        prompts = data_packer.rollout_collate_fn(prompts)

        # List of completions per prompt.
        # [
        #   [completion_str, completion_str, ...],
        #   [completion_str, completion_str, ...],
        #   ...
        # ]
        response: List[List[str]] = []
        try:
            results = self.rollout_engine.generate(
                prompts,
                sampling_params=sampling_params,
                use_tqdm=False,
            )

            for output in results:
                response.append(
                    [output.outputs[i].text for i in range(len(output.outputs))]
                )
        except Exception as e:
            logger.error(f"[Rollout] Failed in rollout generation: {str(e)}")
            import traceback

            traceback.print_exc()
            return []

        return response

    def is_engine_initialized(self):
        return self._engine_initialized

    def get_engine(self):
        if not self._engine_initialized:
            raise RuntimeError(
                "[Rollout] Engine is not initialized, please call init_engine first."
            )
        return self.rollout_engine



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This script is launched by slurm job.
"""

import argparse
import logging
from typing import List
import subprocess
import time
import os
import sys

sys.path.append(os.path.dirname(os.path.realpath(__file__)))
from util import NodeLaunchMetadata


logging.basicConfig(level=logging.INFO)

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--type",
        type=str,
        required=True,
        choices=["policy", "rollout"],
    )
    parser.add_argument(
        "--config",
        type=str,
        default=None,
        help="Path to the config file for the policy or rollout script.",
    )
    parser.add_argument("--script", type=str)
    args = parser.parse_args()

    node_list = os.environ["LOCAL_NODE_LIST"].split(" ")
    COSMOS_CONTROLLER_HOST = os.environ["COSMOS_CONTROLLER_HOST"]
    self_node = os.environ["SLURMD_NODENAME"]
    node_launch_metadata: List[NodeLaunchMetadata] = NodeLaunchMetadata.from_json_list(
        os.environ[f"NODE_LAUNCH_METADATA_{args.type.upper()}"]
    )

    # Either the policy or rollout nodes
    logging.info(f"COSMOS_CONTROLLER_HOST: {COSMOS_CONTROLLER_HOST}")
    logging.info(f"NODE LIST: {node_list}")
    assert self_node in node_list, f"self_node {self_node} not in node_list {node_list}"
    self_node_idx = node_list.index(self_node)
    self_node_launch_metadata = node_launch_metadata[self_node_idx]

    cmds = []
    envs = []
    for replica_launch_metadata in self_node_launch_metadata.colocation:
        rendezvous_node = node_list[replica_launch_metadata.rendezvous_node]
        rendezvous_port = replica_launch_metadata.rendezvous_port
        visible_gpus = replica_launch_metadata.visible_gpus
        nnode = replica_launch_metadata.nnode

        logging.info(
            f"Rendezvous node: {rendezvous_node}, rendezvous port: {rendezvous_port}, visible GPUs: {visible_gpus}"
        )
        env = os.environ.copy()
        env["CUDA_VISIBLE_DEVICES"] = ",".join(str(gpu) for gpu in visible_gpus)
        env["VLLM_DISABLE_COMPILE_CACHE"] = "1"
        replica_launch_script = os.path.join(
            os.path.dirname(__file__),
            "..",
            "..",
            "launcher",
            "launch_replica.sh",
        )
        cmd = [
            replica_launch_script,
            "--type",
            args.type,
            "--rdzv-endpoint",
            f"{rendezvous_node}:{rendezvous_port}",
            "--ngpus",
            str(len(visible_gpus)),
            "--nnodes",
            str(nnode),
        ]
        if args.config:
            cmd += ["--config", args.config]
        if args.script:
            cmd += ["--script", args.script]
        cmds.append(cmd)
        envs.append(env)

    procs = [subprocess.Popen(cmd, env=env) for cmd, env in zip(cmds, envs)]
    # block until every process finishes, and propagate any non-zero exit codes
    exit_code = 0
    while len(procs) > 0:
        for i, p in enumerate(procs):
            try:
                # Check if process has finished without blocking
                if p.poll() is not None:
                    returncode = p.returncode
                    if returncode != 0:
                        # Kill and exit if any process fails
                        for p in procs:
                            try:
                                p.kill()
                            except Exception:
                                pass
                        sys.exit(returncode)
                    # Remove completed process from list
                    procs.remove(p)
            except Exception:
                # Terminate all remaining processes
                for p in procs:
                    try:
                        p.kill()
                    except Exception:
                        pass
                sys.exit(1)
        # Small sleep to prevent busy waiting
        time.sleep(0.1)
    sys.exit(exit_code)  # mimic “all-good” or the first failure



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import redis
from datetime import datetime
from cosmos_rl.utils.constant import (
    RedisStreamConstant,
    COSMOS_HTTP_RETRY_CONFIG,
    COSMOS_HTTP_LONG_WAIT_MAX_RETRY,
)
from typing import List
from cosmos_rl.utils.network_util import make_request_with_retry
from functools import partial
from cosmos_rl.utils.logging import logger
import enum


class RedisOpType(enum.Enum):
    XADD = "add"
    XREAD = "read"
    PING = "ping"


class RedisStreamHandler:
    def __init__(self, ips: List[str], port: int):
        """
        Initialize the RedisStreamHandler.

        Args:
            ips (List[str]): The alternative IP addresses of the Redis server.
            port (int): The port of the Redis server.
            stream_name (str): The name of the Redis stream to interact with.
        """
        self.ips = ips
        self.port = port
        self.redis_clients = []
        for ip in ips:
            self.redis_clients.append(
                redis.Redis(host=ip, port=self.port, db=0, decode_responses=False)
            )
        self.latest_id_command = "0-0"
        self.latest_id_rollout = "0-0"
        self.ping()

    def publish_command(self, data, stream_name: str):
        """
        Write data to the Redis stream.

        Args:
            data : The packed command to write to the stream.

        Returns:
            str: The ID of the added stream entry.
        """
        message = {"command": data, "timestamp": datetime.now().isoformat()}
        # Add message to stream
        try:
            make_request_with_retry(
                self.requests_for_alternative_clients(
                    RedisOpType.XADD,
                    stream_name + "_command",
                    message,
                    maxlen=RedisStreamConstant.STREAM_MAXLEN,
                ),
                response_parser=None,
                max_retries=COSMOS_HTTP_RETRY_CONFIG.max_retries,
            )
        except Exception as e:
            logger.error(f"Failed to write to Redis stream {stream_name}_command: {e}")
            raise e

    def subscribe_command(self, stream_name: str) -> List[dict]:
        """
        Read data from the Redis stream.

        Args:
            stream_name (str): The name of the Redis stream to read from.

        Returns:
            list: A list of stream entries.
        """
        try:
            messages = make_request_with_retry(
                self.requests_for_alternative_clients(
                    RedisOpType.XREAD,
                    {stream_name + "_command": self.latest_id_command},
                    count=RedisStreamConstant.CMD_FETCH_SIZE,
                    block=RedisStreamConstant.CMD_READING_TIMEOUT_MS,
                ),
                response_parser=None,
                max_retries=COSMOS_HTTP_RETRY_CONFIG.max_retries,
            )
        except Exception as e:
            logger.error(f"Failed to read from Redis stream {stream_name}_command: {e}")
            raise e
        commands = []
        if messages:
            for _, message_list in messages:
                for message_id, message_data in message_list:
                    commands.append(message_data[b"command"])
                    self.latest_id_command = message_id
        return commands

    def publish_rollout(self, data, stream_name: str):
        """
        Write data to the Redis stream.

        Args:
            data : The packed rollout to write to the stream.

        Returns:
            str: The ID of the added stream entry.
        """
        message = {"rollout": data, "timestamp": datetime.now().isoformat()}
        # Add message to stream
        try:
            make_request_with_retry(
                self.requests_for_alternative_clients(
                    RedisOpType.XADD,
                    stream_name + "_rollout",
                    message,
                    maxlen=RedisStreamConstant.STREAM_MAXLEN,
                ),
                response_parser=None,
                max_retries=COSMOS_HTTP_RETRY_CONFIG.max_retries,
            )
        except Exception as e:
            logger.error(f"Failed to write to Redis stream {stream_name}_rollout: {e}")
            raise e

    def subscribe_rollout(self, stream_name: str, count: int = -1) -> List[dict]:
        """
        Read data from the Redis stream.

        Args:
            stream_name (str): The name of the Redis stream to read from.
            count (int): The number of messages to read.

        Returns:
            list: A list of stream entries.
        """
        try:
            messages = make_request_with_retry(
                self.requests_for_alternative_clients(
                    RedisOpType.XREAD,
                    {stream_name + "_rollout": self.latest_id_rollout},
                    count=RedisStreamConstant.ROLLOUT_FETCH_SIZE
                    if count <= 0
                    else count,
                    block=RedisStreamConstant.ROLLOUT_READING_TIMEOUT_MS,
                ),
                response_parser=None,
                max_retries=COSMOS_HTTP_RETRY_CONFIG.max_retries,
            )
        except Exception as e:
            logger.error(f"Failed to read from Redis stream {stream_name}_rollout: {e}")
        rollouts = []
        if messages:
            for _, message_list in messages:
                for message_id, message_data in message_list:
                    rollouts.append(message_data[b"rollout"])
                    self.latest_id_rollout = message_id
        return rollouts

    def requests_for_alternative_clients(self, op: RedisOpType, *args, **kwargs):
        """
        Make requests to alternative clients based on the operation type.

        Args:
            op (RedisOpType): The operation type (XADD or XREAD or PING).
            *args: Positional arguments for the request.
            **kwargs: Keyword arguments for the request.

        Returns:
            list: A list of Callable objects for the requests.
        """
        calls = []
        if op == RedisOpType.XADD:
            for redis_client in self.redis_clients:
                calls.append(
                    partial(
                        redis_client.xadd,
                        *args,
                        **kwargs,
                    )
                )
        elif op == RedisOpType.XREAD:
            for redis_client in self.redis_clients:
                calls.append(
                    partial(
                        redis_client.xread,
                        *args,
                        **kwargs,
                    )
                )
        elif op == RedisOpType.PING:
            for redis_client in self.redis_clients:
                calls.append(redis_client.ping)
        else:
            raise ValueError(f"Unsupported operation type: {op}")
        return calls

    def ping(self):
        # Wait for redis to be ready
        try:
            make_request_with_retry(
                self.requests_for_alternative_clients(RedisOpType.PING),
                response_parser=None,
                max_retries=COSMOS_HTTP_LONG_WAIT_MAX_RETRY,
            )
        except Exception as e:
            logger.error(f"Failed to ping Redis when init Redis: {e}")
            raise e
